from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class DescripcionesEspecificas(XElement):
    """
    http://www.sat.gob.mx/ComercioExterior
    Nodo opcional que indica la lista de descripciones específicas de la mercancía. Una mercancía puede tener más de una descripción específica.
    """
    def __init__(
            self,
            marca: str,
            modelo: str = None,
            sub_modelo: str = None,
            numero_serie: str = None,
    ): 
        """
        Nodo opcional que indica la lista de descripciones específicas de la mercancía. Una mercancía puede tener más de una descripción específica.
        
        :param marca: Atributo requerido que indica la marca de la mercancía.
        :param modelo: Atributo opcional que indica el modelo de la mercancía.
        :param sub_modelo: Atributo opcional que indica el submodelo de la mercancía.
        :param numero_serie: Atributo opcional que indica el número de serie de la mercancía.
        """
        
        super().__init__({
            'Marca': marca,
            'Modelo': modelo,
            'SubModelo': sub_modelo,
            'NumeroSerie': numero_serie,
        })
        

class Mercancia(XElement):
    """
    http://www.sat.gob.mx/ComercioExterior
    Nodo requerido para capturar la información de la declaración de cada mercancía exportada.
    """
    def __init__(
            self,
            no_identificacion: str,
            valor_dolares: Decimal | int,
            fraccion_arancelaria: str = None,
            cantidad_aduana: Decimal | int = None,
            unidad_aduana: str = None,
            valor_unitario_aduana: Decimal | int = None,
            descripciones_especificas: DescripcionesEspecificas | dict | Sequence[DescripcionesEspecificas | dict] = None,
    ): 
        """
        Nodo requerido para capturar la información de la declaración de cada mercancía exportada.
        
        :param no_identificacion: Atributo requerido que sirve para expresar el número de parte, la clave de identificación que asigna la empresa o el número de serie de la mercancía exportada.
        :param valor_dolares: Atributo requerido que indica el valor total en dólares de Estados Unidos.
        :param fraccion_arancelaria: Atributo opcional que sirve para expresar la fracción arancelaria correspondiente a la descripción de la mercancía exportada, este dato se vuelve requerido cuando se cuente con él o se esté obligado legalmente a contar con él.
        :param cantidad_aduana: Atributo opcional para precisar la cantidad de bienes en la aduana conforme a la UnidadAduana cuando en el nodo Comprobante:Conceptos:Concepto se hubiera registrado información comercial.
        :param unidad_aduana: Atributo opcional para precisar la unidad de medida aplicable para la cantidad expresada en la mercancía en la aduana.
        :param valor_unitario_aduana: Atributo opcional para precisar el valor o precio unitario del bien en la aduana. Se expresa en dólares de Estados Unidos (USD).
        :param descripciones_especificas: Nodo opcional que indica la lista de descripciones específicas de la mercancía. Una mercancía puede tener más de una descripción específica.
        """
        
        super().__init__({
            'NoIdentificacion': no_identificacion,
            'ValorDolares': valor_dolares,
            'FraccionArancelaria': fraccion_arancelaria,
            'CantidadAduana': cantidad_aduana,
            'UnidadAduana': unidad_aduana,
            'ValorUnitarioAduana': valor_unitario_aduana,
            'DescripcionesEspecificas': descripciones_especificas,
        })
        

class Domicilio(XElement):
    """
    http://www.sat.gob.mx/ComercioExterior
    Nodo requerido para expresar el domicilio del destinatario de la mercancía.
    """
    def __init__(
            self,
            calle: str,
            estado: str,
            pais: str,
            codigo_postal: str,
            numero_exterior: str = None,
            numero_interior: str = None,
            colonia: str = None,
            localidad: str = None,
            referencia: str = None,
            municipio: str = None,
    ): 
        """
        Nodo requerido para expresar el domicilio del destinatario de la mercancía.
        
        :param calle: Atributo requerido sirve para precisar la calle en que está ubicado el domicilio del destinatario de la mercancía.
        :param estado: Atributo requerido para señalar el estado, entidad, región, comunidad u otra figura análoga en donde se encuentra ubicado el domicilio del destinatario de la mercancía. El catálogo se publicará en el portal del SAT y será conforme con la especificación ISO 3166-2.
        :param pais: Atributo requerido que sirve para precisar el país donde se encuentra ubicado el destinatario de la mercancía. El catálogo se publicará en el portal del SAT y será conforme con la especificación ISO 3166-1.
        :param codigo_postal: Atributo requerido que sirve para asentar el código postal (PO, BOX) en donde se encuentra ubicado el domicilio del destinatario de la mercancía.
        :param numero_exterior: Atributo opcional sirve para expresar el número exterior en donde se ubica el domicilio del destinatario de la mercancía.
        :param numero_interior: Atributo opcional sirve para expresar el número interior, en caso de existir, en donde se ubica el domicilio del destinatario de la mercancía.
        :param colonia: Atributo opcional sirve para expresar la colonia o dato análogo en donde se ubica el domicilio del destinatario de la mercancía.
        :param localidad: Atributo opcional que sirve para precisar la ciudad, población, distrito u otro análogo en donde se ubica el domicilio del destinatario de la mercancía.
        :param referencia: Atributo opcional para expresar una referencia geográfica adicional que permita una más fácil o precisa ubicación del domicilio del destinatario de la mercancía, por ejemplo las coordenadas GPS.
        :param municipio: Atributo opcional que sirve para precisar el municipio, delegación, condado u otro análogo en donde se encuentra ubicado el destinatario de la mercancía.
        """
        
        super().__init__({
            'Calle': calle,
            'Estado': estado,
            'Pais': pais,
            'CodigoPostal': codigo_postal,
            'NumeroExterior': numero_exterior,
            'NumeroInterior': numero_interior,
            'Colonia': colonia,
            'Localidad': localidad,
            'Referencia': referencia,
            'Municipio': municipio,
        })
        

class Destinatario(XElement):
    """
    http://www.sat.gob.mx/ComercioExterior
    Nodo opcional para capturar los datos del destinatario de la mercancía cuando éste sea distinto del receptor del CFDI.
    """
    def __init__(
            self,
            domicilio: Domicilio | dict,
            num_reg_id_trib: str = None,
            rfc: str = None,
            curp: str = None,
            nombre: str = None,
    ): 
        """
        Nodo opcional para capturar los datos del destinatario de la mercancía cuando éste sea distinto del receptor del CFDI.
        
        :param domicilio: Nodo requerido para expresar el domicilio del destinatario de la mercancía.
        :param num_reg_id_trib: Atributo opcional para incorporar el número de identificación o registro fiscal del país de residencia para efectos fiscales del destinatario de la mercancía exportada.
        :param rfc: Atributo opcional para expresar el RFC del destinatario de la mercancía exportada.
        :param curp: Atributo opcional para expresar la CURP del destinatario de la mercancía cuando es persona física.
        :param nombre: Atributo opcional para expresar el nombre completo, denominación o razón social del destinatario de la mercancía exportada.
        """
        
        super().__init__({
            'Domicilio': domicilio,
            'NumRegIdTrib': num_reg_id_trib,
            'Rfc': rfc,
            'Curp': curp,
            'Nombre': nombre,
        })
        

class Emisor(XElement):
    """
    http://www.sat.gob.mx/ComercioExterior
    Nodo opcional para capturar los datos complementarios del emisor del comprobante.
    """
    def __init__(
            self,
            curp: str = None,
    ): 
        """
        Nodo opcional para capturar los datos complementarios del emisor del comprobante.
        
        :param curp: Atributo opcional para expresar la CURP del emisor del CFDI cuando es una persona física.
        """
        
        super().__init__({
            'Curp': curp,
        })
        

class Receptor(XElement):
    """
    http://www.sat.gob.mx/ComercioExterior
    Nodo requerido para capturar los datos complementarios del receptor del CFDI.
    """
    def __init__(
            self,
            num_reg_id_trib: str,
            curp: str = None,
    ): 
        """
        Nodo requerido para capturar los datos complementarios del receptor del CFDI.
        
        :param num_reg_id_trib: Atributo requerido para incorporar el número de identificación o registro fiscal del país de residencia para efectos fiscales del receptor del CFDI.
        :param curp: Atributo opcional para expresar la CURP del receptor del CFDI cuando es una persona física.
        """
        
        super().__init__({
            'NumRegIdTrib': num_reg_id_trib,
            'Curp': curp,
        })
        

class ComercioExterior(CFDI):
    """
    Complemento para incorporar la información en el caso de Exportación de Mercancías en definitiva.
    """
    tag = '{http://www.sat.gob.mx/ComercioExterior}ComercioExterior'
    version = '1.0'
    
    def __init__(
            self,
            tipo_operacion: str,
            receptor: Receptor | dict,
            clave_de_pedimento: str = None,
            certificado_origen: int = None,
            num_certificado_origen: str = None,
            numero_exportador_confiable: str = None,
            incoterm: str = None,
            subdivision: int = None,
            observaciones: str = None,
            tipo_cambio_usd: Decimal | int = None,
            total_usd: Decimal | int = None,
            emisor: Emisor | dict = None,
            destinatario: Destinatario | dict = None,
            mercancias: Mercancia | dict | Sequence[Mercancia | dict] = None,
    ): 
        """
        Complemento para incorporar la información en el caso de Exportación de Mercancías en definitiva.
        
        :param tipo_operacion: Atributo requerido que indica el tipo de operación de comercio exterior que se realiza, puede ser importación o exportación, A = exportación de servicios. 2 = exportación.
        :param receptor: Nodo requerido para capturar los datos complementarios del receptor del CFDI.
        :param clave_de_pedimento: Atributo que indica la clave de pedimento que se haya declarado conforme al apéndice 2 del anexo 22 de las reglas generales de comercio exterior.
        :param certificado_origen: Atributo derivado de la excepción de certificados de Origen de los Tratados de Libre Comercio que ha celebrado México con diversos países. 0 = No Funge como certificado de origen 1 = Funge como certificado de origen.
        :param num_certificado_origen: Atributo opcional para expresar el folio del certificado de origen o el folio fiscal del CFDI con el que se pagó la expedición del certificado de origen.
        :param numero_exportador_confiable: Atributo opcional que indica el número de exportador confiable, conforme al artículo 22 del Anexo 1 del Tratado de Libre Comercio con la Asociación Europea y a la Decisión de la Comunidad Europea.
        :param incoterm: Atributo que indica la clave del INCOTERM aplicable a la factura.
        :param subdivision: Atributo que indica si la factura tiene o no subdivisión. Valores posibles:0 - no tiene subdivisión,1 - si tiene subdivisión.
        :param observaciones: Atributo opcional en caso de ingresar alguna información adicional, como alguna leyenda que debe incluir el CFDI.
        :param tipo_cambio_usd: Atributo que indica el número de pesos mexicanos que equivalen a un dólar de Estados Unidos, de acuerdo al artículo 20 del Código Fiscal de la Federación.
        :param total_usd: Atributo que indica el importe total del comprobante en dólares de Estados Unidos.
        :param emisor: Nodo opcional para capturar los datos complementarios del emisor del comprobante.
        :param destinatario: Nodo opcional para capturar los datos del destinatario de la mercancía cuando éste sea distinto del receptor del CFDI.
        :param mercancias: Nodo opcional para capturar la información de la declaración de las mercancías exportadas.
        """
        
        super().__init__({
            'Version': self.version,
            'TipoOperacion': tipo_operacion,
            'Receptor': receptor,
            'ClaveDePedimento': clave_de_pedimento,
            'CertificadoOrigen': certificado_origen,
            'NumCertificadoOrigen': num_certificado_origen,
            'NumeroExportadorConfiable': numero_exportador_confiable,
            'Incoterm': incoterm,
            'Subdivision': subdivision,
            'Observaciones': observaciones,
            'TipoCambioUSD': tipo_cambio_usd,
            'TotalUSD': total_usd,
            'Emisor': emisor,
            'Destinatario': destinatario,
            'Mercancias': mercancias,
        })
        





__title__ = "SAT-CFDI"
__package__ = "satcfdi"

__description__ = "The best open-source python library to generate and process SAT's CFDI"
__git_project__ = "SAT-CFDI/python-satcfdi"
__url__ = f"https://github.com/{__git_project__}"

__docs_url__ = f"https://{__package__}.readthedocs.io/"
__docs_badge_url__ = f"https://readthedocs.org/projects/{__package__}/badge/"

__version__ = "1.0.6"
__author__ = "satcfdi@outlook.com"
__author_email__ = "satcfdi@outlook.com"
__license__ = "MIT License"

__user_agent__ = f"python-{__package__}/{__version__}"





# -*- coding: utf-8 -*-
import base64
from enum import Enum, auto

from OpenSSL import crypto
from OpenSSL.crypto import X509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, rsa

from .curp import CURP
from .rfc import RFC, RFCType
from ..exceptions import CFDIError

REGIMEN_SOCIETARIOS = {
    'AC': 'AC - Asociación civil',
    'S DE CV': 'Sociedad de capital variable',
    'S DE RL DE CV': 'Sociedad de responsabilidad limitada de capital variable',
    'S DE RL': 'Sociedad de responsabilidad limitada',
    'S EN C DE CV': 'Sociedad en comandita simple de capital variable',
    'S EN C POR A DE CV': 'Sociedad en comandita por acciones de capital variable',
    'S EN C POR A': 'Sociedad en comandita por acciones',
    'S EN C': 'Sociedad en comandita simple',
    'S EN NC DE CV': 'Sociedad en nombre colectivo de capital variable',
    'S EN NC': 'Sociedad en nombre colectivo',
    'SA DE CV': 'Sociedad anónima de capital variable',
    'SA': 'Sociedad anónima',
    'SAB DE CV': 'Sociedad anómina bursátil de capital variable',
    'SAB': 'Sociedad anómina bursátil',
    'SAPI DE CV': 'Sociedad anónima promotora de inversión de capital variable',
    'SAPI': 'Sociedad anónima promotora de inversión',
    'SAPIB': 'Sociedad anónima promotora de inversión bursátil',
    'SAS DE CV': 'Sociedad por acciones simplificada de capital variable',
    'SAS': 'Sociedad por acciones simplificada',
    'SC DE CV': 'Sociedad civil de capital variable',
    'SC': 'Sociedad civil',
}


class CertificateType(Enum):
    Fiel = auto()
    CSD = auto()


class Certificate:
    def __init__(self, certificate: X509):
        self.certificate = certificate

    @classmethod
    def load_certificate(cls, certificate: bytes) -> 'Certificate':
        return cls(crypto.load_certificate(crypto.FILETYPE_ASN1, certificate))

    def fingerprint(self, algorithm=hashes.SHA1()) -> bytes:
        return self.certificate.to_cryptography().fingerprint(algorithm=algorithm)

    def certificate_bytes(self) -> bytes:
        return crypto.dump_certificate(crypto.FILETYPE_ASN1, self.certificate)

    def certificate_base64(self) -> str:
        return base64.b64encode(
            self.certificate_bytes()
        ).decode()

    def issuer(self) -> str:
        # return self.certificate.to_cryptography().issuer.rfc4514_string()
        d = self.certificate.get_issuer().get_components()
        return ','.join(f'{k.decode()}={v.decode()}' for k, v in reversed(d))

    def subject(self) -> str:
        d = self.certificate.get_subject().get_components()
        return ','.join(f'{k.decode()}={v.decode()}' for k, v in reversed(d))

    @property
    def type(self):
        if self.certificate.get_extension_count() == 4:
            return CertificateType.Fiel
        if self.certificate.get_extension_count() == 2:
            return CertificateType.CSD
        return None

    @property
    def serial_number(self) -> int:
        return self.certificate.get_serial_number()

    # useful data
    @property
    def legal_name(self) -> str | None:
        try:
            at = self.certificate.get_subject().O
            if self.rfc.type == RFCType.MORAL:
                for r in REGIMEN_SOCIETARIOS:
                    if at.endswith(" " + r):
                        return at[:-len(r) - 1]
            return at
        except AttributeError:
            return None

    @property
    def rfc(self) -> RFC | None:
        try:
            at = self.certificate.get_subject().x500UniqueIdentifier
            at = at.split("/")[0].strip()
            return RFC(at)
        except AttributeError:
            return None

    @property
    def rfc_representante(self) -> RFC | None:
        try:
            at = self.certificate.get_subject().x500UniqueIdentifier
            at = at.split("/")[1].strip()
            return RFC(at)
        except AttributeError:
            return None

    @property
    def curp(self) -> CURP | None:
        try:
            at = self.certificate.get_subject().serialNumber
            at = at.split("/")[0].strip()
            return CURP(at) if at else None
        except AttributeError:
            return None

    @property
    def curp_representante(self) -> CURP | None:
        try:
            at = self.certificate.get_subject().serialNumber
            at = at.split("/")[1].strip()
            return CURP(at)
        except AttributeError:
            return None

    @property
    def branch_name(self) -> str | None:
        try:
            return self.certificate.get_subject().OU
        except AttributeError:
            return None

    @property
    def rfc_pac(self) -> RFC:
        if nom_suc := self.branch_name:
            if nom_suc.startswith("PAC"):
                return RFC(nom_suc[3:15])
            if nom_suc.startswith("MEGAPAC"):
                return RFC(nom_suc[7:19])
        raise CFDIError("Certificado no es PAC")

    @property
    def email(self) -> str | None:
        try:
            return self.certificate.get_subject().emailAddress
        except AttributeError:
            return None

    @property
    def certificate_number(self) -> str:
        return f'{self.certificate.get_serial_number():x}'[1::2]

    def public_key(self) -> rsa.RSAPublicKey:
        return self.certificate.get_pubkey().to_cryptography_key()

    def _verify(self, data, signature, algorithm) -> bool:
        try:
            self.public_key().verify(
                signature=signature,
                data=data,
                padding=padding.PKCS1v15(),
                algorithm=algorithm
            )
            return True
        except InvalidSignature:
            return False

    def verify_sha1(self, data: bytes, signature: bytes) -> bool:
        return self._verify(
            data=data,
            signature=signature,
            algorithm=hashes.SHA1()
        )

    def verify_sha256(self, data: bytes, signature: bytes) -> bool:
        return self._verify(
            data=data,
            signature=signature,
            algorithm=hashes.SHA256()
        )

    def encrypt(self, data: bytes):
        return self.public_key().encrypt(
            plaintext=data,
            padding=padding.PKCS1v15()
        )



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class AportacionesODepositos(XElement):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1/planesderetiro11
    Nodo para identificar el tipo de aportación o depósito que realiza el aportante a su plan de retiro, conforme a los artículos 151, 185 de la Ley el Impuesto sobre la Renta, y 258 del Reglamento de la Ley del Impuesto Sobre la Renta.
    """
    def __init__(
            self,
            tipo_aportacion_odeposito: str,
            mont_aport_odep: Decimal | int,
            rfc_fiduciaria: str = None,
    ): 
        """
        Nodo para identificar el tipo de aportación o depósito que realiza el aportante a su plan de retiro, conforme a los artículos 151, 185 de la Ley el Impuesto sobre la Renta, y 258 del Reglamento de la Ley del Impuesto Sobre la Renta.
        
        :param tipo_aportacion_odeposito: Atributo requerido que identifica el tipo de aportación o depósito que se realiza al plan de retiro.
        :param mont_aport_odep: Atributo requerido que expresa el monto de las aportaciones o depósitos efectuados al plan de retiro.
        :param rfc_fiduciaria: Atributo opcional que expresa el RFC de la fiduciaria de fungió como administrador del plan de pensiones, cuando se trate de aportaciones realizadas de conformidad con el artículo 258 del Reglamento de la LISR, se vuelve requerido cuando el administrador del plan es una fiduciaria.
        """
        
        super().__init__({
            'TipoAportacionODeposito': tipo_aportacion_odeposito,
            'MontAportODep': mont_aport_odep,
            'RFCFiduciaria': rfc_fiduciaria,
        })
        

class Planesderetiro(CFDI):
    """
    Complemento para expresar la información sobre los planes personales de retiro
    """
    tag = '{http://www.sat.gob.mx/esquemas/retencionpago/1/planesderetiro11}Planesderetiro'
    version = '1.1'
    
    def __init__(
            self,
            sistema_financ: str,
            mont_int_reales_deveng_anioo_inm_ant: Decimal | int,
            hubo_retiros_anio_inm_ant_per: str,
            hubo_retiros_anio_inm_ant: str,
            mont_tot_aport_anio_inm_anterior: Decimal | int = None,
            mont_tot_retirado_anio_inm_ant_per: Decimal | int = None,
            mont_tot_exent_retirado_anio_inm_ant: Decimal | int = None,
            mont_tot_exedente_anio_inm_ant: Decimal | int = None,
            mont_tot_retirado_anio_inm_ant: Decimal | int = None,
            num_referencia: str = None,
            aportaciones_odepositos: AportacionesODepositos | dict | Sequence[AportacionesODepositos | dict] = None,
    ): 
        """
        Complemento para expresar la información sobre los planes personales de retiro
        
        :param sistema_financ: Atributo requerido para expresar si los planes personales de retiro son del sistema financiero
        :param mont_int_reales_deveng_anioo_inm_ant: Atributo requerido para expresar el monto de los intereses reales devengados o percibidos durante el año inmediato anterior de los planes personales de retiro
        :param hubo_retiros_anio_inm_ant_per: Atributo requerido para expresar si se realizaron retiros de recursos invertidos y sus rendimientos en el ejercicio inmediato anterior antes de cumplir los requisitos de permanencia
        :param hubo_retiros_anio_inm_ant: Atributo requerido que expresa si se realizaron retiros en el ejercicio inmediato anterior
        :param mont_tot_aport_anio_inm_anterior: Atributo opcional que expresa el monto total de las aportaciones actualizadas en el año inmediato anterior de los planes personales de retiro
        :param mont_tot_retirado_anio_inm_ant_per: Atributo condicional que expresa el monto total del retiro realizado antes de cumplir con los requisitos de permanencia
        :param mont_tot_exent_retirado_anio_inm_ant: Atributo condicional que expresa el monto total exento del retiro realizado en el ejercicio inmediato anterior
        :param mont_tot_exedente_anio_inm_ant: Atributo condicional que expresa el monto total excedente del monto exento del retiro realizado en el ejercicio inmediato anterior
        :param mont_tot_retirado_anio_inm_ant: Atributo opcional que expresa el monto total de retiros realizados en el ejercicio inmediato anterior
        :param num_referencia: Atributo opcional que expresa el número de referencia, contrato o cuenta con la que identifica la institución a su cliente.
        :param aportaciones_odepositos: Nodo para identificar el tipo de aportación o depósito que realiza el aportante a su plan de retiro, conforme a los artículos 151, 185 de la Ley el Impuesto sobre la Renta, y 258 del Reglamento de la Ley del Impuesto Sobre la Renta.
        """
        
        super().__init__({
            'Version': self.version,
            'SistemaFinanc': sistema_financ,
            'MontIntRealesDevengAniooInmAnt': mont_int_reales_deveng_anioo_inm_ant,
            'HuboRetirosAnioInmAntPer': hubo_retiros_anio_inm_ant_per,
            'HuboRetirosAnioInmAnt': hubo_retiros_anio_inm_ant,
            'MontTotAportAnioInmAnterior': mont_tot_aport_anio_inm_anterior,
            'MontTotRetiradoAnioInmAntPer': mont_tot_retirado_anio_inm_ant_per,
            'MontTotExentRetiradoAnioInmAnt': mont_tot_exent_retirado_anio_inm_ant,
            'MontTotExedenteAnioInmAnt': mont_tot_exedente_anio_inm_ant,
            'MontTotRetiradoAnioInmAnt': mont_tot_retirado_anio_inm_ant,
            'NumReferencia': num_referencia,
            'AportacionesODepositos': aportaciones_odepositos,
        })
        




from OpenSSL.crypto import FILETYPE_ASN1
from cryptography import x509
from OpenSSL import crypto


class CertificateSigningRequest:
    def __init__(self, request: x509.CertificateSigningRequest):
        self.request = request

    @classmethod
    def load(cls, request: bytes) -> 'CertificateSigningRequest':
        return cls(x509.load_der_x509_csr(request))

    @classmethod
    def load2(cls, request: bytes):
        return crypto.load_certificate_request(FILETYPE_ASN1, request)



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class SectorFinanciero(CFDI):
    """
    Complemento requerido para uso exclusivo de las entidades integrantes del sistema financiero que actúen en su carácter de fiduciarias
    """
    tag = '{http://www.sat.gob.mx/esquemas/retencionpago/1/sectorfinanciero}SectorFinanciero'
    version = '1.0'
    
    def __init__(
            self,
            id_fideicom: str,
            descrip_fideicom: str,
            nom_fideicom: str = None,
    ): 
        """
        Complemento requerido para uso exclusivo de las entidades integrantes del sistema financiero que actúen en su carácter de fiduciarias
        
        :param id_fideicom: Atributo requerido para expresar el Identificador o Número del Fideicomiso
        :param descrip_fideicom: Atributo requerido para expresar el objeto o fin del Fideicomiso
        :param nom_fideicom: Atributo opcional para expresar el Nombre del Fideicomiso
        """
        
        super().__init__({
            'Version': self.version,
            'IdFideicom': id_fideicom,
            'DescripFideicom': descrip_fideicom,
            'NomFideicom': nom_fideicom,
        })
        




import os
from io import BytesIO

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from satcfdi import RFC, RFCType, Code

BUFFER_SIZE = 65536


def encrypt_triple_des(data: BytesIO):
    pl = 8 - data.tell() % 8
    data.write(bytes([pl for _ in range(pl)]))
    data.seek(0)

    key = os.urandom(24)
    iv = os.urandom(8)

    cipher = Cipher(
        algorithm=algorithms.TripleDES(key),
        mode=modes.CBC(iv)
    )
    encryptor = cipher.encryptor()

    out = BytesIO()
    while True:
        buf = data.read(BUFFER_SIZE)
        if buf:
            out.write(encryptor.update(buf))
        else:
            out.write(encryptor.finalize())
            break

    return "1.2.840.113549.3.7", key, iv, out


class DIOTWriter:
    def __init__(self, stream):
        self.io = stream

    def __call__(self, clave, b, c, valor):
        self.io.write(
            f"{clave}|{b}|{c}|{valor}|\r\n".encode('windows-1252')
        )

    def end(self):
        self.io.write(b"EOF\r\n")


def _format_rfc(rfc: RFC):
    if rfc.type == RFCType.FISICA:
        return str(rfc)
    return " " + rfc


def period_code(period: str):
    period = int(period)
    if period <= 12:
        return "1"
    if period <= 18:
        return "2"
    if period <= 22:
        return "3"
    if period <= 25:
        return "4"
    return "5"


def catalog_code(catalog, value):
    if value is None:
        return None
    return Code(value, catalog[value])



from collections.abc import *
from datetime import datetime, timedelta, timezone
from decimal import Decimal

from .. import Issuer
from ...cfdi import CFDI, XElement


class ImpRetenidos(XElement):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1
    Nodo opcional para expresar el total de los impuestos retenidos que se desprenden de los conceptos expresados en el documento de retenciones e información de pagos.
    """
    def __init__(
            self,
            monto_ret: Decimal | int,
            tipo_pago_ret: str,
            base_ret: Decimal | int = None,
            impuesto: str = None,
    ): 
        """
        Nodo opcional para expresar el total de los impuestos retenidos que se desprenden de los conceptos expresados en el documento de retenciones e información de pagos.
        
        :param monto_ret: Atributo requerido para expresar el importe del impuesto retenido en el periodo o ejercicio
        :param tipo_pago_ret: Atributo requerido para precisar si el monto de la retención es considerado pago definitivo o pago provisional
        :param base_ret: Atributo opcional para expresar la  base del impuesto, que puede ser la diferencia entre los ingresos percibidos y las deducciones autorizadas
        :param impuesto: Atributo opcional para señalar el tipo de impuesto retenido del periodo o ejercicio conforme al catálogo.
        """
        
        super().__init__({
            'MontoRet': monto_ret,
            'TipoPagoRet': tipo_pago_ret,
            'BaseRet': base_ret,
            'Impuesto': impuesto,
        })
        

class Totales(XElement):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1
    Nodo requerido para expresar el total de las retenciones e información de pagos efectuados en el período que ampara el documento.
    """
    def __init__(
            self,
            monto_tot_operacion: Decimal | int,
            monto_tot_grav: Decimal | int,
            monto_tot_exent: Decimal | int,
            monto_tot_ret: Decimal | int,
            imp_retenidos: Sequence[ImpRetenidos | dict] = None,
    ): 
        """
        Nodo requerido para expresar el total de las retenciones e información de pagos efectuados en el período que ampara el documento.
        
        :param monto_tot_operacion: Atributo requerido para expresar  el total del monto de la operación  que se relaciona en el comprobante 
        :param monto_tot_grav: Atributo requerido para expresar el total del monto gravado de la operación  que se relaciona en el comprobante.
        :param monto_tot_exent: Atributo requerido para expresar el total del monto exento de la operación  que se relaciona en el comprobante.
        :param monto_tot_ret: Atributo requerido para expresar el monto total de las retenciones. Sumatoria de los montos de retención del nodo ImpRetenidos.
        :param imp_retenidos: Nodo opcional para expresar el total de los impuestos retenidos que se desprenden de los conceptos expresados en el documento de retenciones e información de pagos.
        """
        
        super().__init__({
            'MontoTotOperacion': monto_tot_operacion,
            'MontoTotGrav': monto_tot_grav,
            'MontoTotExent': monto_tot_exent,
            'MontoTotRet': monto_tot_ret,
            'ImpRetenidos': imp_retenidos,
        })
        

class Periodo(XElement):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1
    Nodo requerido para expresar el periodo que ampara el documento de retenciones e información de pagos
    """
    def __init__(
            self,
            mes_ini: int,
            mes_fin: int,
            ejerc: int,
    ): 
        """
        Nodo requerido para expresar el periodo que ampara el documento de retenciones e información de pagos
        
        :param mes_ini: Atributo requerido para la expresión del mes inicial del periodo de la retención e información de pagos
        :param mes_fin: Atributo requerido para la expresión del mes final del periodo de la retención e información de pagos
        :param ejerc: Atributo requerido para la expresión del ejercicio fiscal (año) 
        """
        
        super().__init__({
            'MesIni': mes_ini,
            'MesFin': mes_fin,
            'Ejerc': ejerc,
        })
        

class Extranjero(XElement):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1
    Nodo requerido para expresar la información del contribuyente receptor del documento cuando sea residente en el extranjero
    """
    def __init__(
            self,
            nom_den_raz_soc_r: str,
            num_reg_id_trib: str = None,
    ): 
        """
        Nodo requerido para expresar la información del contribuyente receptor del documento cuando sea residente en el extranjero
        
        :param nom_den_raz_soc_r: Atributo requerido para expresar el nombre, denominación o razón social del receptor del documento cuando sea residente en el extranjero
        :param num_reg_id_trib: Atributo opcional para expresar el número de registro de identificación fiscal del receptor del documento cuando sea residente en el extranjero
        """
        
        super().__init__({
            'NomDenRazSocR': nom_den_raz_soc_r,
            'NumRegIdTrib': num_reg_id_trib,
        })
        

class Nacional(XElement):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1
    Nodo requerido para expresar la información del contribuyente receptor en caso de que sea de nacionalidad mexicana
    """
    def __init__(
            self,
            rfcrecep: str,
            nom_den_raz_soc_r: str = None,
            curpr: str = None,
    ): 
        """
        Nodo requerido para expresar la información del contribuyente receptor en caso de que sea de nacionalidad mexicana
        
        :param rfcrecep: Atributo requerido para la clave del Registro Federal de Contribuyentes correspondiente al contribuyente receptor del documento.
        :param nom_den_raz_soc_r: Atributo opcional para el nombre, denominación o razón social del contribuyente receptor del documento.
        :param curpr: Atributo opcional para la Clave Única del Registro Poblacional del contribuyente receptor del documento.
        """
        
        super().__init__({
            'RFCRecep': rfcrecep,
            'NomDenRazSocR': nom_den_raz_soc_r,
            'CURPR': curpr,
        })
        

class Receptor(XElement):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1
    Nodo requerido para expresar la información del contribuyente receptor del documento electrónico de retenciones e información de pagos.
    """
    def __init__(
            self,
            nacionalidad: str,
            nacional: Nacional | dict = None,
            extranjero: Extranjero | dict = None,
    ): 
        """
        Nodo requerido para expresar la información del contribuyente receptor del documento electrónico de retenciones e información de pagos.
        
        :param nacionalidad: Atributo requerido para expresar la nacionalidad del receptor del documento.
        :param nacional: Nodo requerido para expresar la información del contribuyente receptor en caso de que sea de nacionalidad mexicana
        :param extranjero: Nodo requerido para expresar la información del contribuyente receptor del documento cuando sea residente en el extranjero
        """
        
        super().__init__({
            'Nacionalidad': nacionalidad,
            'Nacional': nacional,
            'Extranjero': extranjero,
        })


class Retenciones(CFDI):
    """
    Estándar de Documento Electrónico Retenciones e Información de Pagos.
    """
    tag = '{http://www.sat.gob.mx/esquemas/retencionpago/1}Retenciones'
    version = '1.0'

    def __init__(
            self,
            cve_retenc: str,
            emisor: Issuer | dict,
            receptor: Receptor | dict,
            periodo: Periodo | dict,
            totales: Totales | dict,
            folio_int: str = None,
            desc_retenc: str = None,
            complemento: XElement | Sequence[XElement] = None,
            addenda: XElement | Sequence[XElement] = None,
            fecha_exp: datetime = None,
    ):
        """
        Estándar de Documento Electrónico Retenciones e Información de Pagos.

        :param sello: Atributo requerido para contener el sello digital del documento de retención e información de pagos. El sello deberá ser expresado como una cadena de texto en formato base 64.
        :param num_cert: Atributo requerido para expresar el número de serie del certificado de sello digital con el que se selló digitalmente el documento de la retención e información de pagos.
        :param cert: Atributo requerido que sirve para incorporar el certificado de sello digital que ampara el documento de retención e información de pagos como texto, en formato base 64.
        :param fecha_exp: Atributo requerido para la expresión de la fecha y hora de expedición del documento de retención e información de pagos. Se expresa en la forma yyyy-mm-ddThh:mm:ssTZD-6, de acuerdo con la especificación ISO 8601.
        :param cve_retenc: Atributo requerido para expresar la clave de la retención e información de pagos de acuerdo al catálogo publicado en internet por el SAT.
        :param emisor: Nodo requerido para expresar la información del contribuyente emisor del documento electrónico de retenciones e información de pagos.
        :param receptor: Nodo requerido para expresar la información del contribuyente receptor del documento electrónico de retenciones e información de pagos.
        :param periodo: Nodo requerido para expresar el periodo que ampara el documento de retenciones e información de pagos
        :param totales: Nodo requerido para expresar el total de las retenciones e información de pagos efectuados en el período que ampara el documento.
        :param folio_int: Atributo opcional para control interno del contribuyente que expresa el folio del documento que ampara la retención e información de pagos. Permite números y/o letras.
        :param desc_retenc: Atributo opcional que expresa la descripción de la retención e información de pagos en caso de que en el atributo CveRetenc se haya elegido el valor para 'otro tipo de retenciones'
        :param complemento: Nodo opcional donde se incluirá el complemento Timbre Fiscal Digital de manera obligatoria y los nodos complementarios determinados por el SAT, de acuerdo a las disposiciones particulares a un sector o actividad específica.
        :param addenda: Nodo opcional para recibir las extensiones al formato que sean de utilidad al contribuyente. Para las reglas de uso del mismo, referirse al formato de origen.
        """

        fecha_exp = fecha_exp or datetime.now(tz=timezone(timedelta(hours=-6)))
        super().__init__({
            'Version': self.version,
            'Sello': '',
            'NumCert': emisor.certificate_number,
            'Cert':  emisor.signer.certificate_base64(),
            'FechaExp': fecha_exp,
            'CveRetenc': cve_retenc,
            'FolioInt': folio_int,
            'DescRetenc': desc_retenc,
            'Emisor': {
                "RFCEmisor": emisor.rfc,
                "NomDenRazSocE": emisor.legal_name,
                "CURPE": None
            },
            'Receptor': receptor,
            'Periodo': periodo,
            'Totales': totales,
            'Complemento': complemento,
            'Addenda': addenda,
        })
        if emisor.signer:
            self['Sello'] = emisor.signer.sign_sha1(
                self.cadena_original().encode()
            )



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import *

from ... import Signer
from ...cfdi import CFDI, XElement
from ...transform import MEXICO_TZ


class TimbreFiscalDigital(CFDI):
    """
    Complemento requerido para el Timbrado Fiscal Digital que da valides a un Comprobante Fiscal Digital.
    """
    tag = '{http://www.sat.gob.mx/TimbreFiscalDigital}TimbreFiscalDigital'
    version = '1.0'
    
    def __init__(
            self,
            proveedor: Signer,
            uuid: str,
            fecha_timbrado: datetime,
            sello_cfd: str
    ):
        """
        Complemento requerido para el Timbrado Fiscal Digital que da valides a un Comprobante Fiscal Digital.
        
        :param uuid: Atributo requerido para expresar los 36 caracteres del UUID de la transacción de timbrado
        :param fecha_timbrado: Atributo requerido para expresar la fecha y hora de la generación del timbre
        :param sello_cfd: Atributo requerido para contener el sello digital del comprobante fiscal, que será timbrado. El sello deberá ser expresado cómo una cadena de texto en formato Base 64.
        :param no_certificado_sat: Atributo requerido para expresar el número de serie del certificado del SAT usado para el Timbre
        :param sello_sat: Atributo requerido para contener el sello digital del Timbre Fiscal Digital, al que hacen referencia las reglas de resolución miscelánea aplicable. El sello deberá ser expresado cómo una cadena de texto en formato Base 64.
        :return: objeto CFDI
        """

        super().__init__({
            'Version': self.version,
            'UUID': uuid,
            'FechaTimbrado': fecha_timbrado or datetime.now(tz=MEXICO_TZ).replace(tzinfo=None),
            'RfcProvCertif': proveedor.rfc_pac,
            'SelloCFD': sello_cfd,
            'NoCertificadoSAT': proveedor.certificate_number,
            'SelloSAT': '',
        })

        self['SelloSAT'] = proveedor.sign_sha1(
            self.cadena_original().encode()
        )



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class EnajenaciondeAcciones(CFDI):
    """
    Complemento para expresar la enajenación de acciones u operaciones de valores (incluye ganancia o pérdida).
    """
    tag = '{http://www.sat.gob.mx/esquemas/retencionpago/1/enajenaciondeacciones}EnajenaciondeAcciones'
    version = '1.0'
    
    def __init__(
            self,
            contrato_intermediacion: str,
            ganancia: Decimal | int,
            perdida: Decimal | int,
    ): 
        """
        Complemento para expresar la enajenación de acciones u operaciones de valores (incluye ganancia o pérdida).
        
        :param contrato_intermediacion: Atributo requerido para expresar la descripción del contrato de intermediación
        :param ganancia: Atributo requerido para expresar la ganancia obtenida por la enajenación de acciones u operación de valores
        :param perdida: Atributo requerido para expresar la pérdida en el contrato de intermediación
        """
        
        super().__init__({
            'Version': self.version,
            'ContratoIntermediacion': contrato_intermediacion,
            'Ganancia': ganancia,
            'Perdida': perdida,
        })
        




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class DatosComiteTecnicoType(XElement):
    def __init__(
            self,
            comite_tecnico: str,
            modificacion_comite_tecnico: str,
    ): 
        """
        
        :param comite_tecnico: 
        :param modificacion_comite_tecnico: 
        """
        
        super().__init__({
            'ComiteTecnico': comite_tecnico,
            'ModificacionComiteTecnico': modificacion_comite_tecnico,
        })
        

class DatosPropietarioType(XElement):
    def __init__(
            self,
            propietario_solicita: str,
            tipo_persona: TipoPersonaSimpleType | dict = None,
    ): 
        """
        
        :param propietario_solicita: 
        :param tipo_persona: 
        """
        
        super().__init__({
            'PropietarioSolicita': propietario_solicita,
            'TipoPersona': tipo_persona,
        })
        

class AvaluoType(XElement):
    def __init__(
            self,
            tipo_bien: str,
            valor_avaluo: str,
            datos_propietario: DatosPropietarioType | dict,
            descripcion: str = None,
    ): 
        """
        
        :param tipo_bien: 
        :param valor_avaluo: 
        :param datos_propietario: 
        :param descripcion: 
        """
        
        super().__init__({
            'TipoBien': tipo_bien,
            'ValorAvaluo': valor_avaluo,
            'DatosPropietario': datos_propietario,
            'Descripcion': descripcion,
        })
        

class DatosLiquidacionSimpleType(XElement):
    def __init__(
            self,
            moneda: str,
            monto_operacion: str,
    ): 
        """
        
        :param moneda: 
        :param monto_operacion: 
        """
        
        super().__init__({
            'Moneda': moneda,
            'MontoOperacion': monto_operacion,
        })
        

class PersonaMoralGaranteType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            fecha_constitucion: date = None,
            rfc: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param fecha_constitucion: 
        :param rfc: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
        })
        

class PersonaFisicaGaranteType(XElement):
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        """
        
        :param nombre: 
        :param apellido_paterno: 
        :param apellido_materno: 
        :param fecha_nacimiento: 
        :param rfc: 
        :param curp: 
        """
        
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class TipoPersonaGaranteType(XElement):
    def __init__(
            self,
            persona_fisica: PersonaFisicaGaranteType | dict = None,
            persona_moral: PersonaMoralGaranteType | dict = None,
            fideicomiso: FideicomisoType | dict = None,
    ): 
        """
        
        :param persona_fisica: 
        :param persona_moral: 
        :param fideicomiso: 
        """
        
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class DatosOtroType(XElement):
    def __init__(
            self,
            descripcion_garantia: str,
    ): 
        """
        
        :param descripcion_garantia: 
        """
        
        super().__init__({
            'DescripcionGarantia': descripcion_garantia,
        })
        

class DatosInmuebleType(XElement):
    def __init__(
            self,
            tipo_inmueble: str,
            valor_referencia: str,
            codigo_postal: str,
            folio_real: str,
    ): 
        """
        
        :param tipo_inmueble: 
        :param valor_referencia: 
        :param codigo_postal: 
        :param folio_real: 
        """
        
        super().__init__({
            'TipoInmueble': tipo_inmueble,
            'ValorReferencia': valor_referencia,
            'CodigoPostal': codigo_postal,
            'FolioReal': folio_real,
        })
        

class DatosBienMutuoType(XElement):
    def __init__(
            self,
            datos_inmueble: DatosInmuebleType | dict = None,
            datos_otro: DatosOtroType | dict = None,
    ): 
        """
        
        :param datos_inmueble: 
        :param datos_otro: 
        """
        
        super().__init__({
            'DatosInmueble': datos_inmueble,
            'DatosOtro': datos_otro,
        })
        

class DatosGarantiaType(XElement):
    def __init__(
            self,
            tipo_garantia: str,
            datos_bien_mutuo: DatosBienMutuoType | dict = None,
            tipo_persona: TipoPersonaGaranteType | dict = None,
    ): 
        """
        
        :param tipo_garantia: 
        :param datos_bien_mutuo: 
        :param tipo_persona: 
        """
        
        super().__init__({
            'TipoGarantia': tipo_garantia,
            'DatosBienMutuo': datos_bien_mutuo,
            'TipoPersona': tipo_persona,
        })
        

class DatosDeudorType(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaType | dict,
    ): 
        """
        
        :param tipo_persona: 
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
        })
        

class DatosAcreedorType(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaType | dict,
    ): 
        """
        
        :param tipo_persona: 
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
        })
        

class ContratoMutuoCredType(XElement):
    def __init__(
            self,
            tipo_otorgamiento: str,
            datos_acreedor: DatosAcreedorType | dict | Sequence[DatosAcreedorType | dict],
            datos_deudor: DatosDeudorType | dict | Sequence[DatosDeudorType | dict],
            datos_liquidacion: DatosLiquidacionSimpleType | dict | Sequence[DatosLiquidacionSimpleType | dict],
            datos_garantia: DatosGarantiaType | dict | Sequence[DatosGarantiaType | dict] = None,
    ): 
        """
        
        :param tipo_otorgamiento: 
        :param datos_acreedor: 
        :param datos_deudor: 
        :param datos_liquidacion: 
        :param datos_garantia: 
        """
        
        super().__init__({
            'TipoOtorgamiento': tipo_otorgamiento,
            'DatosAcreedor': datos_acreedor,
            'DatosDeudor': datos_deudor,
            'DatosLiquidacion': datos_liquidacion,
            'DatosGarantia': datos_garantia,
        })
        

class DatosCesionType(XElement):
    def __init__(
            self,
            monto_cesion: str,
    ): 
        """
        
        :param monto_cesion: 
        """
        
        super().__init__({
            'MontoCesion': monto_cesion,
        })
        

class DatosCesionarioType(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaType | dict,
    ): 
        """
        
        :param tipo_persona: 
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
        })
        

class DatosCedenteType(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaType | dict,
    ): 
        """
        
        :param tipo_persona: 
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
        })
        

class CesionDerechosType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            tipo_cesion: str,
            datos_cedente: DatosCedenteType | dict,
            datos_cesionario: DatosCesionarioType | dict,
            datos_cesion: DatosCesionType | dict,
            identificador_fideicomiso: str = None,
            rfc: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param tipo_cesion: 
        :param datos_cedente: 
        :param datos_cesionario: 
        :param datos_cesion: 
        :param identificador_fideicomiso: 
        :param rfc: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'TipoCesion': tipo_cesion,
            'DatosCedente': datos_cedente,
            'DatosCesionario': datos_cesionario,
            'DatosCesion': datos_cesion,
            'IdentificadorFideicomiso': identificador_fideicomiso,
            'Rfc': rfc,
        })
        

class DatosFideicomisarioType(XElement):
    def __init__(
            self,
            datos_fideicomisarios_determinados: str,
            tipo_persona: TipoPersona1Type | dict | Sequence[TipoPersona1Type | dict] = None,
    ): 
        """
        
        :param datos_fideicomisarios_determinados: 
        :param tipo_persona: 
        """
        
        super().__init__({
            'DatosFideicomisariosDeterminados': datos_fideicomisarios_determinados,
            'TipoPersona': tipo_persona,
        })
        

class PatrimonioOtroBienType(XElement):
    def __init__(
            self,
            descripcion: str,
            valor_bien: str,
    ): 
        """
        
        :param descripcion: 
        :param valor_bien: 
        """
        
        super().__init__({
            'Descripcion': descripcion,
            'ValorBien': valor_bien,
        })
        

class PatrimonioInmuebleType(XElement):
    def __init__(
            self,
            tipo_inmueble: str,
            codigo_postal: str,
            folio_real: str,
            importe_garantia: str,
    ): 
        """
        
        :param tipo_inmueble: 
        :param codigo_postal: 
        :param folio_real: 
        :param importe_garantia: 
        """
        
        super().__init__({
            'TipoInmueble': tipo_inmueble,
            'CodigoPostal': codigo_postal,
            'FolioReal': folio_real,
            'ImporteGarantia': importe_garantia,
        })
        

class PatrimonioMonetariaType(XElement):
    def __init__(
            self,
            moneda: str,
            monto_operacion: str,
    ): 
        """
        
        :param moneda: 
        :param monto_operacion: 
        """
        
        super().__init__({
            'Moneda': moneda,
            'MontoOperacion': monto_operacion,
        })
        

class DatosTipoPatrimonioType(XElement):
    def __init__(
            self,
            patrimonio_monetario: PatrimonioMonetariaType | dict = None,
            patrimonio_inmueble: PatrimonioInmuebleType | dict = None,
            patrimonio_otro_bien: PatrimonioOtroBienType | dict = None,
    ): 
        """
        
        :param patrimonio_monetario: 
        :param patrimonio_inmueble: 
        :param patrimonio_otro_bien: 
        """
        
        super().__init__({
            'PatrimonioMonetario': patrimonio_monetario,
            'PatrimonioInmueble': patrimonio_inmueble,
            'PatrimonioOtroBien': patrimonio_otro_bien,
        })
        

class DatosFideicomitenteType(XElement):
    def __init__(
            self,
            tipo_movimiento_fideicomitente: str,
            tipo_persona: TipoPersonaType | dict,
            datos_tipo_patrimonio: DatosTipoPatrimonioType | dict | Sequence[DatosTipoPatrimonioType | dict] = None,
    ): 
        """
        
        :param tipo_movimiento_fideicomitente: 
        :param tipo_persona: 
        :param datos_tipo_patrimonio: 
        """
        
        super().__init__({
            'TipoMovimientoFideicomitente': tipo_movimiento_fideicomitente,
            'TipoPersona': tipo_persona,
            'DatosTipoPatrimonio': datos_tipo_patrimonio,
        })
        

class ConstModFideicomisoType(XElement):
    def __init__(
            self,
            tipo_movimiento: str,
            tipo_fideicomiso: str,
            denominacion_razon: str,
            monto_patrimonio: str,
            datos_fideicomitente: DatosFideicomitenteType | dict | Sequence[DatosFideicomitenteType | dict],
            descripcion: str = None,
            rfc: str = None,
            identificador_fideicomiso: str = None,
            datos_fideicomisarios: DatosFideicomisarioType | dict | Sequence[DatosFideicomisarioType | dict] = None,
            datos_miembro_comite_tecnico: DatosComiteTecnicoType | dict = None,
    ): 
        """
        
        :param tipo_movimiento: 
        :param tipo_fideicomiso: 
        :param denominacion_razon: 
        :param monto_patrimonio: 
        :param datos_fideicomitente: 
        :param descripcion: 
        :param rfc: 
        :param identificador_fideicomiso: 
        :param datos_fideicomisarios: 
        :param datos_miembro_comite_tecnico: 
        """
        
        super().__init__({
            'TipoMovimiento': tipo_movimiento,
            'TipoFideicomiso': tipo_fideicomiso,
            'DenominacionRazon': denominacion_razon,
            'MontoPatrimonio': monto_patrimonio,
            'DatosFideicomitente': datos_fideicomitente,
            'Descripcion': descripcion,
            'Rfc': rfc,
            'IdentificadorFideicomiso': identificador_fideicomiso,
            'DatosFideicomisarios': datos_fideicomisarios,
            'DatosMiembroComiteTecnico': datos_miembro_comite_tecnico,
        })
        

class DatosLiquidacionType(XElement):
    def __init__(
            self,
            fecha_pago: date,
            moneda: str,
            monto_operacion: str,
            instrumento_monetario: str = None,
    ): 
        """
        
        :param fecha_pago: 
        :param moneda: 
        :param monto_operacion: 
        :param instrumento_monetario: 
        """
        
        super().__init__({
            'FechaPago': fecha_pago,
            'Moneda': moneda,
            'MontoOperacion': monto_operacion,
            'InstrumentoMonetario': instrumento_monetario,
        })
        

class DatosCompradorType(XElement):
    def __init__(
            self,
            numero_acciones_compradas: str,
            tipo_persona: TipoPersonaSimpleType | dict,
    ): 
        """
        
        :param numero_acciones_compradas: 
        :param tipo_persona: 
        """
        
        super().__init__({
            'NumeroAccionesCompradas': numero_acciones_compradas,
            'TipoPersona': tipo_persona,
        })
        

class DatosVendedorType(XElement):
    def __init__(
            self,
            numero_acciones_vendidas: str,
            tipo_persona: TipoPersonaSimpleType | dict,
    ): 
        """
        
        :param numero_acciones_vendidas: 
        :param tipo_persona: 
        """
        
        super().__init__({
            'NumeroAccionesVendidas': numero_acciones_vendidas,
            'TipoPersona': tipo_persona,
        })
        

class PersonaMoralAccionesType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            valor_nominal: str,
            numero_acciones: str,
            datos_vendedor: DatosVendedorType | dict | Sequence[DatosVendedorType | dict],
            datos_comprador: DatosCompradorType | dict | Sequence[DatosCompradorType | dict],
            fecha_constitucion: date = None,
            rfc: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param pais_nacionalidad: 
        :param valor_nominal: 
        :param numero_acciones: 
        :param datos_vendedor: 
        :param datos_comprador: 
        :param fecha_constitucion: 
        :param rfc: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'ValorNominal': valor_nominal,
            'NumeroAcciones': numero_acciones,
            'DatosVendedor': datos_vendedor,
            'DatosComprador': datos_comprador,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
        })
        

class CompraVentaAccionesType(XElement):
    def __init__(
            self,
            tipo_operacion: str,
            persona_moral_acciones: PersonaMoralAccionesType | dict | Sequence[PersonaMoralAccionesType | dict],
            datos_liquidacion: DatosLiquidacionType | dict | Sequence[DatosLiquidacionType | dict],
    ): 
        """
        
        :param tipo_operacion: 
        :param persona_moral_acciones: 
        :param datos_liquidacion: 
        """
        
        super().__init__({
            'TipoOperacion': tipo_operacion,
            'PersonaMoralAcciones': persona_moral_acciones,
            'DatosLiquidacion': datos_liquidacion,
        })
        

class EscindidaType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            capital_social_fijo: str,
            numero_total_acciones: str,
            datos_accionista: DatosAccionista1Type | dict | Sequence[DatosAccionista1Type | dict],
            fecha_constitucion: date = None,
            rfc: str = None,
            capital_social_variable: str = None,
            folio_mercantil: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param pais_nacionalidad: 
        :param giro_mercantil: 
        :param capital_social_fijo: 
        :param numero_total_acciones: 
        :param datos_accionista: 
        :param fecha_constitucion: 
        :param rfc: 
        :param capital_social_variable: 
        :param folio_mercantil: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'CapitalSocialFijo': capital_social_fijo,
            'NumeroTotalAcciones': numero_total_acciones,
            'DatosAccionista': datos_accionista,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'CapitalSocialVariable': capital_social_variable,
            'FolioMercantil': folio_mercantil,
        })
        

class DatosEscindidasType(XElement):
    def __init__(
            self,
            escindidas_determinadas: str,
            dato_escindida: EscindidaType | dict | Sequence[EscindidaType | dict] = None,
    ): 
        """
        
        :param escindidas_determinadas: 
        :param dato_escindida: 
        """
        
        super().__init__({
            'EscindidasDeterminadas': escindidas_determinadas,
            'DatoEscindida': dato_escindida,
        })
        

class DatosEscindenteType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            capital_social_fijo: str,
            escindente_subsiste: str,
            fecha_constitucion: date = None,
            rfc: str = None,
            capital_social_variable: str = None,
            folio_mercantil: str = None,
            datos_accionista_escindente: DatosAccionista1Type | dict | Sequence[DatosAccionista1Type | dict] = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param pais_nacionalidad: 
        :param giro_mercantil: 
        :param capital_social_fijo: 
        :param escindente_subsiste: 
        :param fecha_constitucion: 
        :param rfc: 
        :param capital_social_variable: 
        :param folio_mercantil: 
        :param datos_accionista_escindente: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'CapitalSocialFijo': capital_social_fijo,
            'EscindenteSubsiste': escindente_subsiste,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'CapitalSocialVariable': capital_social_variable,
            'FolioMercantil': folio_mercantil,
            'DatosAccionistaEscindente': datos_accionista_escindente,
        })
        

class EscisionType(XElement):
    def __init__(
            self,
            datos_escindente: DatosEscindenteType | dict,
            datos_escindidas: DatosEscindidasType | dict,
    ): 
        """
        
        :param datos_escindente: 
        :param datos_escindidas: 
        """
        
        super().__init__({
            'DatosEscindente': datos_escindente,
            'DatosEscindidas': datos_escindidas,
        })
        

class FusionanteType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            capital_social_fijo: str,
            numero_total_acciones: str,
            datos_accionista: DatosAccionista1Type | dict | Sequence[DatosAccionista1Type | dict],
            fecha_constitucion: date = None,
            rfc: str = None,
            capital_social_variable: str = None,
            folio_mercantil: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param pais_nacionalidad: 
        :param giro_mercantil: 
        :param capital_social_fijo: 
        :param numero_total_acciones: 
        :param datos_accionista: 
        :param fecha_constitucion: 
        :param rfc: 
        :param capital_social_variable: 
        :param folio_mercantil: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'CapitalSocialFijo': capital_social_fijo,
            'NumeroTotalAcciones': numero_total_acciones,
            'DatosAccionista': datos_accionista,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'CapitalSocialVariable': capital_social_variable,
            'FolioMercantil': folio_mercantil,
        })
        

class DatosFusionanteType(XElement):
    def __init__(
            self,
            fusionante_determinadas: str,
            fusionante: FusionanteType | dict = None,
    ): 
        """
        
        :param fusionante_determinadas: 
        :param fusionante: 
        """
        
        super().__init__({
            'FusionanteDeterminadas': fusionante_determinadas,
            'Fusionante': fusionante,
        })
        

class DatosFusionadaType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            capital_social_fijo: str,
            fecha_constitucion: date = None,
            rfc: str = None,
            capital_social_variable: str = None,
            folio_mercantil: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param pais_nacionalidad: 
        :param giro_mercantil: 
        :param capital_social_fijo: 
        :param fecha_constitucion: 
        :param rfc: 
        :param capital_social_variable: 
        :param folio_mercantil: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'CapitalSocialFijo': capital_social_fijo,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'CapitalSocialVariable': capital_social_variable,
            'FolioMercantil': folio_mercantil,
        })
        

class DatosFusionadasType(XElement):
    def __init__(
            self,
            datos_fusionada: DatosFusionadaType | dict | Sequence[DatosFusionadaType | dict],
    ): 
        """
        
        :param datos_fusionada: 
        """
        
        super().__init__({
            'DatosFusionada': datos_fusionada,
        })
        

class FusionType(XElement):
    def __init__(
            self,
            tipo_fusion: str,
            datos_fusionadas: DatosFusionadasType | dict,
            datos_fusionante: DatosFusionanteType | dict,
    ): 
        """
        
        :param tipo_fusion: 
        :param datos_fusionadas: 
        :param datos_fusionante: 
        """
        
        super().__init__({
            'TipoFusion': tipo_fusion,
            'DatosFusionadas': datos_fusionadas,
            'DatosFusionante': datos_fusionante,
        })
        

class DatosAccionista1Type(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaSimpleType | dict,
            numero_acciones: str,
    ): 
        """
        
        :param tipo_persona: 
        :param numero_acciones: 
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
            'NumeroAcciones': numero_acciones,
        })
        

class DatosModificacionType(XElement):
    def __init__(
            self,
            tipo_modificacion_capital_fijo: str,
            inicial_capital_fijo: str,
            final_capital_fijo: str,
            tipo_modificacion_capital_variable: str,
            inicial_capital_variable: str,
            final_capital_variable: str,
            datos_accionista: DatosAccionista1Type | dict | Sequence[DatosAccionista1Type | dict],
    ): 
        """
        
        :param tipo_modificacion_capital_fijo: 
        :param inicial_capital_fijo: 
        :param final_capital_fijo: 
        :param tipo_modificacion_capital_variable: 
        :param inicial_capital_variable: 
        :param final_capital_variable: 
        :param datos_accionista: 
        """
        
        super().__init__({
            'TipoModificacionCapitalFijo': tipo_modificacion_capital_fijo,
            'InicialCapitalFijo': inicial_capital_fijo,
            'FinalCapitalFijo': final_capital_fijo,
            'TipoModificacionCapitalVariable': tipo_modificacion_capital_variable,
            'InicialCapitalVariable': inicial_capital_variable,
            'FinalCapitalVariable': final_capital_variable,
            'DatosAccionista': datos_accionista,
        })
        

class PersonaMoralModificaType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            numero_total_acciones: str,
            motivo_modificacion: str,
            fecha_constitucion: date = None,
            rfc: str = None,
            instrumento_publico: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param pais_nacionalidad: 
        :param giro_mercantil: 
        :param numero_total_acciones: 
        :param motivo_modificacion: 
        :param fecha_constitucion: 
        :param rfc: 
        :param instrumento_publico: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'NumeroTotalAcciones': numero_total_acciones,
            'MotivoModificacion': motivo_modificacion,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'InstrumentoPublico': instrumento_publico,
        })
        

class ModificacionPatrimonialType(XElement):
    def __init__(
            self,
            persona_moral_modifica: PersonaMoralModificaType | dict,
            datos_modificacion: DatosModificacionType | dict,
    ): 
        """
        
        :param persona_moral_modifica: 
        :param datos_modificacion: 
        """
        
        super().__init__({
            'PersonaMoralModifica': persona_moral_modifica,
            'DatosModificacion': datos_modificacion,
        })
        

class CapitalSocialType(XElement):
    def __init__(
            self,
            capital_fijo: str,
            capital_variable: str = None,
    ): 
        """
        
        :param capital_fijo: 
        :param capital_variable: 
        """
        
        super().__init__({
            'CapitalFijo': capital_fijo,
            'CapitalVariable': capital_variable,
        })
        

class DatosAccionistaType(XElement):
    def __init__(
            self,
            cargo_accionista: str,
            tipo_persona: TipoPersonaSimpleType | dict,
            numero_acciones: str,
    ): 
        """
        
        :param cargo_accionista: 
        :param tipo_persona: 
        :param numero_acciones: 
        """
        
        super().__init__({
            'CargoAccionista': cargo_accionista,
            'TipoPersona': tipo_persona,
            'NumeroAcciones': numero_acciones,
        })
        

class ConstitucionPersonasMoralesType(XElement):
    def __init__(
            self,
            tipo_persona_moral: str,
            denominacion_razon: str,
            giro_mercantil: str,
            numero_total_acciones: str,
            entidad_federativa: str,
            consejo_vigilancia: str,
            motivo_constitucion: str,
            datos_accionista: DatosAccionistaType | dict | Sequence[DatosAccionistaType | dict],
            capital_social: CapitalSocialType | dict,
            tipo_persona_moral_otra: str = None,
            folio_mercantil: str = None,
            instrumento_publico: str = None,
    ): 
        """
        
        :param tipo_persona_moral: 
        :param denominacion_razon: 
        :param giro_mercantil: 
        :param numero_total_acciones: 
        :param entidad_federativa: 
        :param consejo_vigilancia: 
        :param motivo_constitucion: 
        :param datos_accionista: 
        :param capital_social: 
        :param tipo_persona_moral_otra: 
        :param folio_mercantil: 
        :param instrumento_publico: 
        """
        
        super().__init__({
            'TipoPersonaMoral': tipo_persona_moral,
            'DenominacionRazon': denominacion_razon,
            'GiroMercantil': giro_mercantil,
            'NumeroTotalAcciones': numero_total_acciones,
            'EntidadFederativa': entidad_federativa,
            'ConsejoVigilancia': consejo_vigilancia,
            'MotivoConstitucion': motivo_constitucion,
            'DatosAccionista': datos_accionista,
            'CapitalSocial': capital_social,
            'TipoPersonaMoralOtra': tipo_persona_moral_otra,
            'FolioMercantil': folio_mercantil,
            'InstrumentoPublico': instrumento_publico,
        })
        

class PersonaFisicaSimpleType(XElement):
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            pais_nacionalidad: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        """
        
        :param nombre: 
        :param apellido_paterno: 
        :param apellido_materno: 
        :param pais_nacionalidad: 
        :param fecha_nacimiento: 
        :param rfc: 
        :param curp: 
        """
        
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'PaisNacionalidad': pais_nacionalidad,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class PersonaMoralSimpleType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            fecha_constitucion: date = None,
            rfc: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param pais_nacionalidad: 
        :param fecha_constitucion: 
        :param rfc: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
        })
        

class TipoPersonaSimpleType(XElement):
    def __init__(
            self,
            persona_fisica: PersonaFisicaSimpleType | dict = None,
            persona_moral: PersonaMoralSimpleType | dict = None,
            fideicomiso: FideicomisoType | dict = None,
    ): 
        """
        
        :param persona_fisica: 
        :param persona_moral: 
        :param fideicomiso: 
        """
        
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class FideicomisoType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            rfc: str = None,
            identificador_fideicomiso: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param rfc: 
        :param identificador_fideicomiso: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'Rfc': rfc,
            'IdentificadorFideicomiso': identificador_fideicomiso,
        })
        

class PersonaMoralType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            fecha_constitucion: date = None,
            rfc: str = None,
    ): 
        """
        
        :param denominacion_razon: 
        :param pais_nacionalidad: 
        :param giro_mercantil: 
        :param fecha_constitucion: 
        :param rfc: 
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
        })
        

class PersonaFisicaType(XElement):
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            pais_nacionalidad: str,
            actividad_economica: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        """
        
        :param nombre: 
        :param apellido_paterno: 
        :param apellido_materno: 
        :param pais_nacionalidad: 
        :param actividad_economica: 
        :param fecha_nacimiento: 
        :param rfc: 
        :param curp: 
        """
        
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'PaisNacionalidad': pais_nacionalidad,
            'ActividadEconomica': actividad_economica,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class TipoPersonaType(XElement):
    def __init__(
            self,
            persona_fisica: PersonaFisicaType | dict = None,
            persona_moral: PersonaMoralType | dict = None,
            fideicomiso: FideicomisoType | dict = None,
    ): 
        """
        
        :param persona_fisica: 
        :param persona_moral: 
        :param fideicomiso: 
        """
        
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class DatosApoderadoType(XElement):
    def __init__(
            self,
            tipo_poder: str,
            tipo_persona: TipoPersonaSimpleType | dict,
    ): 
        """
        
        :param tipo_poder: 
        :param tipo_persona: 
        """
        
        super().__init__({
            'TipoPoder': tipo_poder,
            'TipoPersona': tipo_persona,
        })
        

class DatosPoderdanteType(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaType | dict,
    ): 
        """
        
        :param tipo_persona: 
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
        })
        

class OtorgamientoPoderType(XElement):
    def __init__(
            self,
            datos_poderdante: DatosPoderdanteType | dict | Sequence[DatosPoderdanteType | dict],
            datos_apoderado: DatosApoderadoType | dict | Sequence[DatosApoderadoType | dict],
    ): 
        """
        
        :param datos_poderdante: 
        :param datos_apoderado: 
        """
        
        super().__init__({
            'DatosPoderdante': datos_poderdante,
            'DatosApoderado': datos_apoderado,
        })
        

class TipoActividadType(XElement):
    def __init__(
            self,
            otorgamiento_poder: OtorgamientoPoderType | dict = None,
            constitucion_personas_morales: ConstitucionPersonasMoralesType | dict = None,
            modificacion_patrimonial: ModificacionPatrimonialType | dict = None,
            fusion: FusionType | dict = None,
            escision: EscisionType | dict = None,
            compra_venta_acciones: CompraVentaAccionesType | dict = None,
            constitucion_modificacion_fideicomiso: ConstModFideicomisoType | dict = None,
            cesion_derechos_fideicomitente_fideicomisario: CesionDerechosType | dict = None,
            contrato_mutuo_credito: ContratoMutuoCredType | dict = None,
            avaluo: AvaluoType | dict = None,
    ): 
        """
        
        :param otorgamiento_poder: 
        :param constitucion_personas_morales: 
        :param modificacion_patrimonial: 
        :param fusion: 
        :param escision: 
        :param compra_venta_acciones: 
        :param constitucion_modificacion_fideicomiso: 
        :param cesion_derechos_fideicomitente_fideicomisario: 
        :param contrato_mutuo_credito: 
        :param avaluo: 
        """
        
        super().__init__({
            'OtorgamientoPoder': otorgamiento_poder,
            'ConstitucionPersonasMorales': constitucion_personas_morales,
            'ModificacionPatrimonial': modificacion_patrimonial,
            'Fusion': fusion,
            'Escision': escision,
            'CompraVentaAcciones': compra_venta_acciones,
            'ConstitucionModificacionFideicomiso': constitucion_modificacion_fideicomiso,
            'CesionDerechosFideicomitenteFideicomisario': cesion_derechos_fideicomitente_fideicomisario,
            'ContratoMutuoCredito': contrato_mutuo_credito,
            'Avaluo': avaluo,
        })
        

class DatosOperacionType(XElement):
    def __init__(
            self,
            instrumento_publico: str,
            fecha_operacion: date,
            tipo_actividad: TipoActividadType | dict,
    ): 
        """
        
        :param instrumento_publico: 
        :param fecha_operacion: 
        :param tipo_actividad: 
        """
        
        super().__init__({
            'InstrumentoPublico': instrumento_publico,
            'FechaOperacion': fecha_operacion,
            'TipoActividad': tipo_actividad,
        })
        

class DetalleOperacionesType(XElement):
    def __init__(
            self,
            datos_operacion: DatosOperacionType | dict | Sequence[DatosOperacionType | dict],
    ): 
        """
        
        :param datos_operacion: 
        """
        
        super().__init__({
            'DatosOperacion': datos_operacion,
        })
        

class PersonaAvisoType(XElement):
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        """
        
        :param nombre: 
        :param apellido_paterno: 
        :param apellido_materno: 
        :param fecha_nacimiento: 
        :param rfc: 
        :param curp: 
        """
        
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class AlertaType(XElement):
    def __init__(
            self,
            tipo_alerta: str,
            descripcion_alerta: str = None,
    ): 
        """
        
        :param tipo_alerta: 
        :param descripcion_alerta: 
        """
        
        super().__init__({
            'TipoAlerta': tipo_alerta,
            'DescripcionAlerta': descripcion_alerta,
        })
        

class ModificatorioType(XElement):
    def __init__(
            self,
            folio_modificacion: str,
            descripcion_modificacion: str,
    ): 
        """
        
        :param folio_modificacion: 
        :param descripcion_modificacion: 
        """
        
        super().__init__({
            'FolioModificacion': folio_modificacion,
            'DescripcionModificacion': descripcion_modificacion,
        })
        

class AvisoType(XElement):
    def __init__(
            self,
            referencia_aviso: str,
            prioridad: str,
            alerta: AlertaType | dict,
            persona_aviso: PersonaAvisoType | dict | Sequence[PersonaAvisoType | dict],
            detalle_operaciones: DetalleOperacionesType | dict,
            modificatorio: ModificatorioType | dict = None,
    ): 
        """
        
        :param referencia_aviso: 
        :param prioridad: 
        :param alerta: 
        :param persona_aviso: 
        :param detalle_operaciones: 
        :param modificatorio: 
        """
        
        super().__init__({
            'ReferenciaAviso': referencia_aviso,
            'Prioridad': prioridad,
            'Alerta': alerta,
            'PersonaAviso': persona_aviso,
            'DetalleOperaciones': detalle_operaciones,
            'Modificatorio': modificatorio,
        })
        

class SujetoObligadoType(XElement):
    def __init__(
            self,
            clave_sujeto_obligado: str,
            clave_actividad: str,
            clave_entidad_colegiada: str = None,
            exento: str = None,
    ): 
        """
        
        :param clave_sujeto_obligado: 
        :param clave_actividad: 
        :param clave_entidad_colegiada: 
        :param exento: 
        """
        
        super().__init__({
            'ClaveSujetoObligado': clave_sujeto_obligado,
            'ClaveActividad': clave_actividad,
            'ClaveEntidadColegiada': clave_entidad_colegiada,
            'Exento': exento,
        })
        

class InformeType(XElement):
    def __init__(
            self,
            mes_reportado: str,
            sujeto_obligado: SujetoObligadoType | dict,
            aviso: AvisoType | dict | Sequence[AvisoType | dict] = None,
    ): 
        """
        
        :param mes_reportado: 
        :param sujeto_obligado: 
        :param aviso: 
        """
        
        super().__init__({
            'MesReportado': mes_reportado,
            'SujetoObligado': sujeto_obligado,
            'Aviso': aviso,
        })
        

class ArchivoType(XElement):
    def __init__(
            self,
            informe: InformeType | dict | Sequence[InformeType | dict],
    ): 
        """
        
        :param informe: 
        """
        
        super().__init__({
            'Informe': informe,
        })
        

class TipoPersona1Type(XElement):
    def __init__(
            self,
            tipo_movimiento_fideicomisario: str,
            persona_fisica: PersonaFisicaType | dict = None,
            persona_moral: PersonaMoralType | dict = None,
            fideicomiso: FideicomisoType | dict = None,
    ): 
        """
        
        :param tipo_movimiento_fideicomisario: 
        :param persona_fisica: 
        :param persona_moral: 
        :param fideicomiso: 
        """
        
        super().__init__({
            'TipoMovimientoFideicomisario': tipo_movimiento_fideicomisario,
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class Archivo(ArchivoType):
    tag = '{http://www.uif.shcp.gob.mx/recepcion/fep}archivo'




from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from uuid import UUID

from .. import CFDI, XElement, CFDIError
from ..create.compute import make_impuestos_dr_parcial, rounder, group_impuestos, encode_impuesto, calculate_partial
from ..pacs.sat import SAT
from ..transform.helpers import fmt_decimal, strcode

PPD = "PPD"
PUE = "PUE"

sat = SAT()


class SatCFDI(CFDI):
    relations = None  # type: list[Relation]
    payments = None  # type: list[Payment]
    """
    SatCFDI is an extension of a CFDI to represent a CFDI that has been sent to SAT
    """

    @property
    def uuid(self):
        return UUID(self["Complemento"]["TimbreFiscalDigital"]["UUID"])

    @property
    def name(self):
        return self.get("Serie", "") + self.get("Folio", "")

    @property
    def saldo_pendiente(self) -> Decimal:
        if self["TipoDeComprobante"] == "I":
            # Nota de crédito de los documentos relacionados
            credit_notes = sum(
                c.comprobante["Total"]
                for c in self.relations
                if c.cfdi_relacionados["TipoRelacion"] == "01" and c.comprobante['TipoDeComprobante'] == "E" and c.comprobante.estatus != '0'
            )
            insoluto = min(
                (c.docto_relacionado['ImpSaldoInsoluto']
                 for c in self.payments
                 if c.comprobante.estatus != '0'),
                default=None
            )
            if insoluto is not None:
                return insoluto - credit_notes

            insoluto = self["Total"] - credit_notes
            if self["MetodoPago"] == PPD:
                return insoluto
            if self["MetodoPago"] == PUE:
                return Decimal(0)

        return None

    def consulta_estado(self):
        raise NotImplementedError()

    @property
    def metadata(self) -> str:
        match self.tag:
            case '{http://www.sat.gob.mx/cfd/3}Comprobante' | '{http://www.sat.gob.mx/cfd/4}Comprobante':
                # Uuid~RfcEmisor~NombreEmisor~RfcReceptor~NombreReceptor~RfcPac~FechaEmision~FechaCertificacionSat~Monto~EfectoComprobante~Estatus~FechaCancelacion
                return "{uuid}~{rfc_emisor}~{nombre_emisor}~{rfc_receptor}~{nombre_receptor}~{rfc_pac}~{fecha_emision}~{fecha_certificacion_sat}~{monto}~{efecto_comprobante}~{estatus}~{fecha_cancelacion}".format(
                    uuid=self["Complemento"]["TimbreFiscalDigital"]["UUID"],
                    rfc_emisor=self['Emisor']["Rfc"],
                    nombre_emisor=self['Emisor']['Nombre'],
                    rfc_receptor=self['Receptor']['Rfc'],
                    nombre_receptor=self['Receptor'].get('Nombre', ''),
                    rfc_pac=self["Complemento"]["TimbreFiscalDigital"]["RfcProvCertif"],
                    fecha_emision=self["Fecha"],
                    fecha_certificacion_sat=self["Complemento"]["TimbreFiscalDigital"]["FechaTimbrado"],
                    monto=fmt_decimal(self["Total"]),
                    efecto_comprobante=strcode(self["TipoDeComprobante"]),
                    estatus=self.estatus,
                    fecha_cancelacion=self.fecha_cancelacion or ''
                )

            case '{http://www.sat.gob.mx/esquemas/retencionpago/1}Retenciones':
                # Uuid~RfcEmisor~NombreEmisor~RfcReceptor~NombreReceptor~RfcPac~FechaEmision~FechaCertificacionSat~MontoOp~MontoRet~Estatus~FechaCancelacion
                cert_sat = sat.recover_certificate(
                    no_certificado=self["Complemento"]["TimbreFiscalDigital"]["NoCertificadoSAT"]
                )
                return "{uuid}~{rfc_emisor}~{nombre_emisor}~{rfc_receptor}~{nombre_receptor}~{rfc_pac}~{fecha_emision}~{fecha_certificacion_sat}~{monto_op}~{monto_ret}~{estatus}~{fecha_cancelacion}".format(
                    uuid=self["Complemento"]["TimbreFiscalDigital"]["UUID"],
                    rfc_emisor=self["Emisor"]["RFCEmisor"],
                    nombre_emisor=self['Emisor']['NomDenRazSocE'],
                    rfc_receptor=self['Receptor']['Nacional']['RFCRecep'],
                    nombre_receptor=self['Receptor']['Nacional']['NomDenRazSocR'],
                    rfc_pac=cert_sat.rfc_pac,
                    fecha_emision=self["FechaExp"],
                    fecha_certificacion_sat=self["Complemento"]["TimbreFiscalDigital"]["FechaTimbrado"],
                    monto_op=fmt_decimal(self["Totales"]["MontoTotOperacion"]),
                    monto_ret=fmt_decimal(self["Totales"]["MontoTotRet"]),
                    estatus=self.estatus,
                    fecha_cancelacion=self.fecha_cancelacion or ''
                )

            case '{http://www.sat.gob.mx/esquemas/retencionpago/2}Retenciones':
                # Uuid~RfcEmisor~NombreEmisor~RfcReceptor~NombreReceptor~RfcPac~FechaEmision~FechaCertificacionSat~MontoOp~MontoRet~Estatus~FechaCancelacion
                return "{uuid}~{rfc_emisor}~{nombre_emisor}~{rfc_receptor}~{nombre_receptor}~{rfc_pac}~{fecha_emision}~{fecha_certificacion_sat}~{monto_op}~{monto_ret}~{estatus}~{fecha_cancelacion}".format(
                    uuid=self["Complemento"]["TimbreFiscalDigital"]["UUID"],
                    rfc_emisor=self["Emisor"]["RfcE"],
                    nombre_emisor=self['Emisor']['NomDenRazSocE'],
                    rfc_receptor=self['Receptor']['Nacional']['RfcR'],
                    nombre_receptor=self['Receptor']['Nacional']['NomDenRazSocR'],
                    rfc_pac=self["Complemento"]["TimbreFiscalDigital"]["RfcProvCertif"],
                    fecha_emision=self["FechaExp"],
                    fecha_certificacion_sat=self["Complemento"]["TimbreFiscalDigital"]["FechaTimbrado"],
                    monto_op=fmt_decimal(self["Totales"]["MontoTotOperacion"]),
                    monto_ret=fmt_decimal(self["Totales"]["MontoTotRet"]),
                    estatus=self.estatus,
                    fecha_cancelacion=self.fecha_cancelacion or ''
                )
            case _:
                raise CFDIError("No metadata")

    @property
    def estatus(self) -> str:
        """
        '0' = 'Cancelado'
        '1' = 'Vigente'
        :return:
        """
        raise NotImplementedError()

    @property
    def fecha_cancelacion(self) -> datetime | None:
        raise NotImplementedError()


@dataclass(slots=True, init=True)
class Relation:
    cfdi_relacionados: XElement
    comprobante: SatCFDI


@dataclass(slots=True, init=True)
class Payment:
    comprobante: SatCFDI
    pago: XElement = None
    docto_relacionado: XElement = None


@dataclass
class PaymentsDetails(Payment):
    comprobante_pagado: SatCFDI = None

    def __post_init__(self):
        if self.pago:
            self.impuestos = self.docto_relacionado.get('ImpuestosDR')
            if self.impuestos is None:
                self.impuestos = make_impuestos_dr_parcial(
                    conceptos=self.comprobante_pagado['Conceptos'],
                    imp_saldo_ant=self.docto_relacionado['ImpSaldoAnt'],
                    imp_pagado=self.docto_relacionado['ImpPagado'],
                    total=self.comprobante_pagado['Total'],
                    rnd_fn=rounder(self.comprobante_pagado['Moneda'])
                )

            self.impuestos = group_impuestos([{
                "ImpuestosDR": self.impuestos
            }], pfx="DR", ofx="")

            for imp in ('Traslados', 'Retenciones'):
                if imps := self.impuestos.get(imp):
                    self.impuestos[imp] = {
                        encode_impuesto(
                            impuesto=v['Impuesto'],
                            tipo_factor=v.get("TipoFactor"),
                            tasa_cuota=v.get('TasaOCuota')
                        ): v
                        for v in imps
                    }

            def calc_parcial(field):
                return calculate_partial(
                    value=self.comprobante_pagado.get(field),
                    imp_saldo_ant=self.docto_relacionado['ImpSaldoAnt'],
                    imp_pagado=self.docto_relacionado["ImpPagado"],
                    total=self.comprobante_pagado["Total"],
                    rnd_fn=rounder(self.comprobante_pagado['Moneda'])
                )

            self.sub_total = calc_parcial("SubTotal")
            self.descuento = calc_parcial("Descuento")

        else:
            self.impuestos = self.comprobante.get("Impuestos", {})
            self.sub_total = self.comprobante["SubTotal"]
            self.descuento = self.comprobante.get("Descuento")



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class Donatarias(CFDI):
    """
    Nodo opcional para incluir la información requerida por el Servicio de Administración Tributaria a las organizaciones civiles o fideicomisos autorizados para recibir donativos, que permite hacer deducibles los Comprobantes Fiscales Digitales (CFD) y Comprobantes Fiscales Digitales a través de Internet (CFDI) a los donantes.
    """
    tag = '{http://www.sat.gob.mx/donat}Donatarias'
    version = '1.1'
    
    def __init__(
            self,
            no_autorizacion: str,
            fecha_autorizacion: date,
            leyenda: str,
    ): 
        """
        Nodo opcional para incluir la información requerida por el Servicio de Administración Tributaria a las organizaciones civiles o fideicomisos autorizados para recibir donativos, que permite hacer deducibles los Comprobantes Fiscales Digitales (CFD) y Comprobantes Fiscales Digitales a través de Internet (CFDI) a los donantes.
        
        :param no_autorizacion: Atributo requerido para expresar el número del oficio en que se haya informado a la organización civil o fideicomiso, la procedencia de la autorización para recibir donativos deducibles, o su renovación correspondiente otorgada por el Servicio de Administración Tributaria.
        :param fecha_autorizacion: Atributo requerido para expresar la fecha del oficio en que se haya informado a la organización civil o fideicomiso, la procedencia de la autorización para recibir donativos deducibles, o su renovación correspondiente otorgada por el Servicio de Administración Tributaria.
        :param leyenda: Atributo requerido para señalar de manera expresa que el comprobante que se expide se deriva de un donativo.
        """
        
        super().__init__({
            'Version': self.version,
            'NoAutorizacion': no_autorizacion,
            'FechaAutorizacion': fecha_autorizacion,
            'Leyenda': leyenda,
        })
        




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class Traslado(XElement):
    """
    http://www.sat.gob.mx/ecc
    Nodo para la definición de información detallada de un traslado de impuesto específico
    """
    def __init__(
            self,
            impuesto: str,
            tasa: Decimal | int,
            importe: Decimal | int,
    ): 
        """
        Nodo para la definición de información detallada de un traslado de impuesto específico
        
        :param impuesto: Atributo requerido para definir el tipo de impuesto trasladado
        :param tasa: Atributo requerido para señalar la tasa del impuesto que se traslada por cada concepto amparado en el comprobante
        :param importe: Atributo requerido para definir el importe o monto del impuesto trasladado
        """
        
        super().__init__({
            'Impuesto': impuesto,
            'Tasa': tasa,
            'Importe': importe,
        })
        

class ConceptoEstadoDeCuentaCombustible(XElement):
    """
    http://www.sat.gob.mx/ecc
    Nodo requerido para la expresión de una transacción a ser reportada en el estado de cuenta del proveedor de monedero electrónico para operaciones de compra de combustibles.
    """
    def __init__(
            self,
            identificador: str,
            fecha: datetime,
            rfc: str,
            clave_estacion: str,
            cantidad: Decimal | int,
            nombre_combustible: str,
            folio_operacion: str,
            valor_unitario: Decimal | int,
            importe: Decimal | int,
            traslados: Traslado | dict | Sequence[Traslado | dict],
    ): 
        """
        Nodo requerido para la expresión de una transacción a ser reportada en el estado de cuenta del proveedor de monedero electrónico para operaciones de compra de combustibles.
        
        :param identificador: Atributo requerido para la expresión del identificador o número del monedero electrónico
        :param fecha: Atributo requerido para la expresión de la Fecha y hora de expedición de la operación reportada. Se expresa en la forma aaaa-mm-ddThh:mm:ss, de acuerdo con la especificación ISO 8601.
        :param rfc: Atributo requerido del RFC del enajenante del combustible
        :param clave_estacion: Atributo requerido para expresar la clave de cliente de la estación de servicio, a 10 caracteres, cuando sea requerido.
        :param cantidad: Atributo requerido para definir el volumen de combustible adquirido.
        :param nombre_combustible: Atributo requerido para expresar el nombre del combustible adquirido.
        :param folio_operacion: Atributo requerido para referir el número de folio de cada operación realizada por cada monedero electrónico.
        :param valor_unitario: Atributo requerido para definir el precio unitario del combustible adquirido.
        :param importe: Atributo requerido para definir el monto total de consumo de combustible. Debe ser equivalente al resultado de multiplicar la cantidad por el valor unitario.
        :param traslados: Nodo requerido para enlistar los impuestos trasladados aplicables de combustibles.
        """
        
        super().__init__({
            'Identificador': identificador,
            'Fecha': fecha,
            'Rfc': rfc,
            'ClaveEstacion': clave_estacion,
            'Cantidad': cantidad,
            'NombreCombustible': nombre_combustible,
            'FolioOperacion': folio_operacion,
            'ValorUnitario': valor_unitario,
            'Importe': importe,
            'Traslados': traslados,
        })
        

class EstadoDeCuentaCombustible(CFDI):
    """
    Complemento al Comprobante Fiscal Digital (CFD) y Comprobante Fiscal Digital por Internet (CFDI) para integrar la información aplicable al estado de cuenta emitido por un prestador de servicios de monedero electrónico.
    """
    tag = '{http://www.sat.gob.mx/ecc}EstadoDeCuentaCombustible'
    tipo_operacion = 'Tarjeta'
    
    def __init__(
            self,
            numero_de_cuenta: str,
            total: Decimal | int,
            conceptos: ConceptoEstadoDeCuentaCombustible | dict | Sequence[ConceptoEstadoDeCuentaCombustible | dict],
            sub_total: Decimal | int = None,
    ): 
        """
        Complemento al Comprobante Fiscal Digital (CFD) y Comprobante Fiscal Digital por Internet (CFDI) para integrar la información aplicable al estado de cuenta emitido por un prestador de servicios de monedero electrónico.
        
        :param numero_de_cuenta: Nodo requerido para expresar el número de cuenta del adquirente del monedero electrónico
        :param total: Atributo requerido para expresar el monto total de consumo de combustible.
        :param conceptos: Nodo requerido para enlistar los conceptos cubiertos por Estado de Cuenta de Combustible.
        :param sub_total: Atributo opcional para representar la suma de todos los importes tipo ConceptoEstadoDeCuentaCombustible.
        """
        
        super().__init__({
            'TipoOperacion': self.tipo_operacion,
            'NumeroDeCuenta': numero_de_cuenta,
            'Total': total,
            'Conceptos': conceptos,
            'SubTotal': sub_total,
        })
        




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class Divisas(CFDI):
    """
    Complemento al Comprobante Fiscal Digital (CFD) y Comprobante Fiscal Digital por Internet (CFDI) para identificar las operaciones de compra y venta de divisas que realizan los centros cambiarios y las casa de cambio; haciendo mención expresa de que los comprobantes se expiden por la “compra”, o bien, por la “venta” de divisas.
    """
    tag = '{http://www.sat.gob.mx/divisas}Divisas'
    version = '1.0'
    
    def __init__(
            self,
            tipo_operacion: str,
    ): 
        """
        Complemento al Comprobante Fiscal Digital (CFD) y Comprobante Fiscal Digital por Internet (CFDI) para identificar las operaciones de compra y venta de divisas que realizan los centros cambiarios y las casa de cambio; haciendo mención expresa de que los comprobantes se expiden por la “compra”, o bien, por la “venta” de divisas.
        
        :param tipo_operacion: Elemento para definir el tipo de operación realizada. venta o compra de divisas
        """
        
        super().__init__({
            'Version': self.version,
            'TipoOperacion': tipo_operacion,
        })
        




from ..utils import StrEnum


class TipoOperacion(StrEnum):
    PRESTACION_DE_SERVICIOS_PROFESIONALES = "03"
    ARRENDAMIENTO_DE_INMUEBLES = "06"
    OTROS = "85"


class TipoTercero(StrEnum):
    PROVEEDOR_NACIONAL = "04"
    PROVEEDOR_EXTRANJERO = "05"
    PROVEEDOR_GLOBAL = "15"


class Periodo(StrEnum):
    ENERO = "01"
    FEBRERO = "02"
    MARZO = "03"
    ABRIL = "04"
    MAYO = "05"
    JUNIO = "06"
    JULIO = "07"
    AGOSTO = "08"
    SEPTIEMBRE = "09"
    OCTUBRE = "10"
    NOVIEMBRE = "11"
    DICIEMBRE = "12"

    ENERO_FEBRERO = "13"
    MARZO_ABRIL = "14"
    MAYO_JUNIO = "15"
    JULIO_AGOSTO = "16"
    SEPTIEMBRE_OCTUBRE = "17"
    NOVIEMBRE_DICIEMBRE = "18"

    ENERO_MARZO = "19"
    ABRIL_JUNIO = "20"
    JULIO_SEPTIEMBRE = "21"
    OCTUBRE_DICIEMBRE = "22"

    ENERO_ABRIL = "23"
    MAYO_AGOSTO = "24"
    SEPTIEMBRE_DICIEMBRE = "25"

    ENERO_JUNIO = "26"
    JULIO_DICIEMBRE = "27"


class Pais(StrEnum):
    PRINCIPADO_DE_ANDORRA = 'AD'
    EMIRATOS_ARABES_UNIDOS = 'AE'
    AFGANISTAN = 'AF'
    ANTIGUA_Y_BERMUDA = 'AG'
    ISLA_ANGUILLA = 'AI'
    REPUBLICA_DE_ALBANIA = 'AL'
    ANTILLAS_NEERLANDESAS = 'AN'
    REPUBLICA_DE_ANGOLA = 'AO'
    ANTARTICA = 'AQ'
    ARGENTINA = 'AR'
    SAMOA_AMERICANA = 'AS'
    AUSTRIA = 'AT'
    AUSTRALIA = 'AU'
    ARUBA = 'AW'
    ASCENSION = 'AX'
    ISLAS_AZORES = 'AZ'
    BARBADOS = 'BB'
    BANGLADESH = 'BD'
    BELGICA = 'BE'
    BURKINA_FASO = 'BF'
    BULGARIA = 'BG'
    ESTADO_DE_BAHREIN = 'BH'
    BURUNDI = 'BI'
    BENIN = 'BJ'
    BELICE = 'BL'
    BERMUDAS = 'BM'
    BRUNEI_DARUSSALAM = 'BN'
    BOLIVIA = 'BO'
    BRASIL = 'BR'
    COMMONWEALTH_DE_LAS_BAHAMAS = 'BS'
    BUTHAN = 'BT'
    BURMA = 'BU'
    ISLA_BOUVET = 'BV'
    BOTSWANA = 'BW'
    BIELORRUSIA = 'BY'
    CANADA = 'CA'
    ISLA_DE_COCOS_O_KELLING = 'CC'
    ISLAS_CANARIAS = 'CD'
    ISLA_DE_CHRISTMAS = 'CE'
    REPUBLICA_CENTRO_AFRICANA = 'CF'
    CONGO = 'CG'
    SUIZA = 'CH'
    COSTA_DE_MARFIL = 'CI'
    ISLAS_COOK = 'CK'
    CHILE = 'CL'
    CAMERUN = 'CM'
    CHINA = 'CN'
    COLOMBIA = 'CO'
    CAMPIONE_DITALIA = 'CP'
    REPUBLICA_DE_COSTA_RICA = 'CR'
    REPUBLICA_CHECA_Y_REPUBLICA_ESLOVACA = 'CS'
    CUBA = 'CU'
    REPUBLICA_DE_CABO_VERDE = 'CV'
    ISLA_DE_NAVIDAD = 'CX'
    REPUBLICA_DE_CHIPRE = 'CY'
    ALEMANIA = 'DD'
    REPUBLICA_DE_DJIBOUTI = 'DJ'
    DINAMARCA = 'DK'
    REPUBLICA_DOMINICANA = 'DM'
    COMMONWEALTH_DE_DOMINICA = 'DN'
    ARGELIA = 'DZ'
    ECUADOR = 'EC'
    EGIPTO = 'EG'
    SAHARA_DEL_OESTE = 'EH'
    ESTADO_INDEPENDIENTE_DE_SAMOA_OCCIDENTAL = 'EO'
    ESPANA = 'ES'
    ETIOPIA = 'ET'
    FINLANDIA = 'FI'
    FIJI = 'FJ'
    ISLAS_MALVINAS = 'FK'
    MICRONESIA = 'FM'
    ISLAS_FAROE = 'FO'
    FRANCIA = 'FR'
    GABON = 'GA'
    GRAN_BRETANA_REINO_UNIDO = 'GB'
    GRANADA = 'GD'
    GUYANA_FRANCESA = 'GF'
    GHANA = 'GH'
    GIBRALTAR = 'GI'
    GROENLANDIA = 'GJ'
    GAMBIA = 'GM'
    GUINEA = 'GN'
    GUADALUPE = 'GP'
    GUINEA_ECUATORIAL = 'GQ'
    GRECIA = 'GR'
    GUATEMALA = 'GT'
    GUAM = 'GU'
    GUINEA_BISSAU = 'GW'
    REPUBLICA_DE_GUYANA = 'GY'
    ISLAS_DE_GUERNESEY_JERSEY_ALDERNEY_ISLA_GREAT_SARK_HERM_LITTLE_SARK_BERCHOU_JETHOU_LIHOU_ISLAS_DEL_CANAL = 'GZ'
    HONG_KONG = 'HK'
    ISLAS_HEARD_AND_MC_DONALD = 'HM'
    REPUBLICA_DE_HONDURAS = 'HN'
    HAITI = 'HT'
    HUNGRIA = 'HU'
    INDONESIA = 'ID'
    IRLANDA = 'IE'
    ISLA_DEL_HOMBRE = 'IH'
    ISRAEL = 'IL'
    INDIA = 'IN'
    TERRITORIO_BRITANICO_EN_EL_OCEANO_INDICO = 'IO'
    ISLAS_PACIFICO = 'IP'
    IRAQ = 'IQ'
    IRAN = 'IR'
    ISLANDIA = 'IS'
    ITALIA = 'IT'
    JAMAICA = 'JM'
    REINO_HACHEMITA_DE_JORDANIA = 'JO'
    JAPON = 'JP'
    KENIA = 'KE'
    CAMPUCHEA_DEMOCRATICA = 'KH'
    KIRIBATI = 'KI'
    COMOROS = 'KM'
    SAN_KITTS = 'KN'
    REPUBLICA_DEMOCRATICA_DE_COREA = 'KP'
    REPUBLICA_DE_COREA = 'KR'
    ESTADO_DE_KUWAIT = 'KW'
    ISLAS_CAIMAN = 'KY'
    REPUBLICA_DEMOCRATICA_DE_LAOS = 'LA'
    LIBANO = 'LB'
    SANTA_LUCIA = 'LC'
    PRINCIPADO_DE_LIECHTENSTEIN = 'LI'
    REPUBLICA_SOCIALISTA_DEMOCRATICA_DE_SRI_LANKA = 'LK'
    LABUAN = 'LN'
    REPUBLICA_DE_LIBERIA = 'LR'
    LESOTHO = 'LS'
    GRAN_DUCADO_DE_LUXEMBURGO = 'LU'
    LIBIA = 'LY'
    MARRUECOS = 'MA'
    PRINCIPADO_DE_MONACO = 'MC'
    MADEIRA = 'MD'
    MADAGASCAR = 'MG'
    REPUBLICA_DE_LAS_ISLAS_MARSHALL = 'MH'
    MALI = 'ML'
    MONGOLIA = 'MN'
    MACAO = 'MO'
    ISLAS_MARIANAS_DEL_NORESTE = 'MP'
    MARTINICA = 'MQ'
    MAURITANIA = 'MR'
    MONSERRAT = 'MS'
    MALTA = 'MT'
    REPUBLICA_DE_MAURICIO = 'MU'
    REPUBLICA_DE_MALDIVAS = 'MV'
    MALAWI = 'MW'
    MALASIA_ = 'MY'
    MOZAMBIQUE = 'MZ'
    REPUBLICA_DE_NAMIBIA = 'NA'
    NUEVA_CALEDONIA = 'NC'
    NIGER = 'NE'
    ISLA_DE_NORFOLK = 'NF'
    NIGERIA = 'NG'
    NICARAGUA = 'NI'
    HOLANDA = 'NL'
    NORUEGA = 'NO'
    NEPAL = 'NP'
    REPUBLICA_DE_NAURU = 'NR'
    ZONA_NEUTRAL = 'NT'
    NIUE = 'NU'
    NEVIS = 'NV'
    NUEVA_ZELANDIA = 'NZ'
    SULTANIA_DE_OMAN = 'OM'
    REPUBLICA_DE_PANAMA = 'PA'
    PERU = 'PE'
    POLINESIA_FRANCESA = 'PF'
    PAPUA_NUEVA_GUINEA = 'PG'
    FILIPINAS = 'PH'
    PAKISTAN = 'PK'
    POLONIA = 'PL'
    ISLA_DE_SAN_PEDRO_Y_MIGUELON = 'PM'
    PITCAIRN = 'PN'
    ESTADO_LIBRE_ASOCIADO_DE_PUERTO_RICO = 'PR'
    PORTUGAL = 'PT'
    PATAU = 'PU'
    PALAU = 'PW'
    PARAGUAY = 'PY'
    ESTADO_DE_QUATAR = 'QA'
    ISLA_QESHM = 'QB'
    REUNION = 'RE'
    RUMANIA = 'RO'
    RHUANDA = 'RW'
    ARABIA_SAUDITA = 'SA'
    ISLAS_SALOMON = 'SB'
    SEYCHELLES_ISLAS = 'SC'
    SUDAN = 'SD'
    SUECIA = 'SE'
    SINGAPUR = 'SG'
    SANTA_ELENA = 'SH'
    ARCHIPIELAGO_DE_SVALBARD = 'SI'
    ISLAS_SVALBARD_AND_JAN_MAYEN = 'SJ'
    SARK = 'SK'
    SIERRA_LEONA = 'SL'
    SERENISIMA_REPUBLICA_DE_SAN_MARINO = 'SM'
    SENEGAL = 'SN'
    SOMALIA = 'SO'
    SURINAM = 'SR'
    SAO_TOME_AND_PRINCIPE = 'ST'
    PAISES_DE_LA_EX_URSS_EXCEPTO_UCRANIA_Y_BIELORUSIA = 'SU'
    EL_SALVADOR = 'SV'
    REPUBLICA_DE_SEYCHELLES = 'SW'
    SIRIA = 'SY'
    REINO_DE_SWAZILANDIA = 'SZ'
    ISLAS_TURCAS_Y_CAICOS = 'TC'
    CHAD = 'TD'
    TERRITORIOS_FRANCESES_DEL_SURESTE = 'TF'
    TOGO = 'TG'
    THAILANDIA = 'TH'
    TOKELAU = 'TK'
    REPUBLICA_DE_TUNEZ = 'TN'
    REINO_DE_TONGA = 'TO'
    TIMOR_ESTE = 'TP'
    TRIESTE = 'TR'
    TRISTAN_DA_CUNHA = 'TS'
    REPUBLICA_DE_TRINIDAD_Y_TOBAGO = 'TT'
    TURQUIA = 'TU'
    TUVALU = 'TV'
    TAIWAN = 'TW'
    TANZANIA = 'TZ'
    UCRANIA = 'UA'
    UGANDA = 'UG'
    ISLAS_MENORES_ALEJADAS_DE_LOS_ESTADOS_UNIDOS = 'UM'
    ESTADOS_UNIDOS_DE_AMERICA = 'US'
    REPUBLICA_ORIENTAL_DEL_URUGUAY = 'UY'
    EL_VATICANO = 'VA'
    SAN_VICENTE_Y_LAS_GRANADINAS = 'VC'
    VENEZUELA = 'VE'
    ISLAS_VIRGENES_BRITANICAS = 'VG'
    ISLAS_VIRGENES_DE_ESTADOS_UNIDOS_DE_AMERICA = 'VI'
    VIETNAM = 'VN'
    REPUBLICA_DE_VANUATU = 'VU'
    ISLAS_WALLIS_Y_FUNTUNA = 'WF'
    OTRO = 'XX'
    YEMEN_DEMOCRATICA = 'YD'
    REPUBLICA_DEL_YEMEN = 'YE'
    PAISES_DE_LA_EX_YUGOSLAVIA = 'YU'
    SUDAFRICA = 'ZA'
    ZONA_ESPECIAL_CANARIA = 'ZC'
    ZAMBIA = 'ZM'
    ZONA_LIBRE_OSTRAVA = 'ZO'
    ZAIRE = 'ZR'
    ZIMBAWE = 'ZW'



import logging
from decimal import Decimal
from lxml import etree
from lxml.etree import QName

from .. import CFDIError
from ..models import Code
# noinspection PyUnresolvedReferences
from ..transform.catalog import TRANSLATIONS, CATALOGS
from ..utils import iterate

logger = logging.getLogger(__name__)
TEXT_KEY = "_text"
_ = iterate


class SchemaCollector:
    def __init__(self):
        self.nsmap = {}
        self.schemas = []
        self.base = set()

    def add_schema(self, schema):
        if schema not in self.schemas:
            self.schemas.append(schema)

    def add_map(self, tns_alias, tns):
        if tns_alias in self.nsmap:
            if not self.nsmap[tns_alias] == tns:
                raise CFDIError("TNS Alias already registered with a different alias")
        else:
            self.nsmap[tns_alias] = tns

    def add_base(self, base):
        self.base.add(base)


class Xint(int):
    """
    An int class that knows the original value that was used to create it
    """

    def __new__(cls, value: str):
        self = int.__new__(cls, value)
        self.value = value
        return self

    def __str__(self):
        return self.value


def fmt_decimal(d):
    if isinstance(d, Decimal):
        return d.__format__("f")
    return str(d)


def simple_element(_tag, nsmap=None, text=None):
    el = etree.Element(_tag, nsmap=nsmap)
    el.text = text
    return el


def default_objectify(cls, node):
    self = cls()
    self.tag = node.tag

    def add_attributes(n, target):
        for k, v in n.attrib.items():
            target[k] = v

        text = n.text
        if text:
            text = n.text.strip()
            if text:
                target[TEXT_KEY] = text

    def add_elements(n, target):
        for el in n:
            new = cls()
            name = QName(el.tag).localname
            add_attributes(el, target=new)
            add_elements(el, target=new)

            current = target.get(name)
            if current is None:
                target[name] = new
            elif isinstance(current, cls):
                target[name] = [current, new]
            else:
                current.append(new)

    add_attributes(node, target=self)
    add_elements(node, target=self)

    return self


def impuesto_index(attrib, attribute_name):
    impuesto = attrib[attribute_name]
    ext = attribute_name[8:]
    tipo_factor = attrib.get('TipoFactor' + ext)
    if tipo_factor:
        impuesto += "|" + tipo_factor
        tasa_cuota = attrib.get('TasaOCuota' + ext)
        if tasa_cuota:
            impuesto += "|" + Decimal(tasa_cuota).__format__(".6f")

    return impuesto


def catalog_code(*args):
    desc = CATALOGS
    code = args[1]
    try:
        for arg in args:
            if arg is None:  # like Postal Code
                desc = None
                continue
            desc = desc[arg]
    except KeyError:
        logger.error("Key Not found: %s", " ".join(args))
        desc = None

    return Code(code, desc)


def strcode(data):
    if isinstance(data, Code):
        return data.code
    return data


def format_address_raw(calle, num_exterior, num_interior, referencia, colonia, municipio, localidad, estado, pais, codigo_postal):
    parts = []

    calle_num = None
    if calle or num_exterior or num_interior:
        if num_exterior:
            calle_num = f"{calle} #{num_exterior}"
        else:
            calle_num = f"{calle}"

        if num_interior:
            calle_num = f"{calle_num}, int. #{num_interior}"

    if colonia:
        if calle_num:
            parts.append(f"{calle_num}, {colonia}")
        else:
            parts.append(f"{colonia}")

    if referencia:
        parts.append(f"{referencia}")

    if localidad and localidad != municipio:
        parts.append(f"{localidad}")

    if municipio:
        parts.append(f"{municipio}, {estado} {codigo_postal}")
    else:
        parts.append(f"{estado} {codigo_postal}")

    parts.append(f"{pais}")
    return "\n".join(parts)


def format_address(k):
    return format_address_raw(
        calle=k["Calle"],
        num_exterior=k.get("NumeroExterior"),
        num_interior=k.get("NumeroInterior"),
        referencia=desc(k.get("Referencia")),
        colonia=desc(k.get("Colonia")),
        municipio=desc(k.get("Municipio")),
        localidad=desc(k.get("Localidad")),
        estado=desc(k["Estado"]),
        pais=desc(k["Pais"]),
        codigo_postal=k["CodigoPostal"]
    )


def split_at_upper(word: str):
    def split_at_upper_itr(word: str):
        piu = None
        for w in word:
            niu = w.isupper()
            if piu == False:
                if niu:
                    yield " "

            if piu is None:
                yield w.upper()
            else:
                yield w
            piu = niu

    return "".join(split_at_upper_itr(word))


def trans(k):
    res = TRANSLATIONS.get(k)
    if res is None:
        res = split_at_upper(k)
    return res


def desc(s):
    if isinstance(s, Code):
        return s.description
    return s



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import *

from . import cfdi40, pago10
from .. import Issuer
from ...cfdi import CFDI, XElement

_a = Issuer


class Impuesto(cfdi40.Impuesto):
    """
    http://www.sat.gob.mx/cfd/4
    Nodo requerido para la información detallada de un traslado de impuesto específico.
    """


class CfdiRelacionados(XElement):
    """
    http://www.sat.gob.mx/cfd/3
    Nodo opcional para precisar la información de los comprobantes relacionados.
    """

    def __init__(
            self,
            tipo_relacion: str,
            cfdi_relacionado: str | Sequence[str],
    ):
        """
        Nodo opcional para precisar la información de los comprobantes relacionados.

        :param tipo_relacion: Atributo requerido para indicar la clave de la relación que existe entre éste que se esta generando y el o los CFDI previos.
        :param cfdi_relacionado: Nodo requerido para precisar la información de los comprobantes relacionados.
        """

        super().__init__({
            'TipoRelacion': tipo_relacion,
            'CfdiRelacionado': cfdi_relacionado,
        })


class Parte(XElement):
    """
    http://www.sat.gob.mx/cfd/3
    Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el comprobante fiscal digital por Internet.
    """

    def __init__(
            self,
            clave_prod_serv: str,
            cantidad: Decimal | int,
            descripcion: str,
            no_identificacion: str = None,
            unidad: str = None,
            valor_unitario: Decimal | int = None,
            importe: Decimal | int = None,
            informacion_aduanera: str | Sequence[str] = None,
    ):
        """
        Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el comprobante fiscal digital por Internet.

        :param clave_prod_serv: Atributo requerido para expresar la clave del producto o del servicio amparado por la presente parte. Es requerido y deben utilizar las claves del catálogo de productos y servicios, cuando los conceptos que registren por sus actividades correspondan con dichos conceptos.
        :param cantidad: Atributo requerido para precisar la cantidad de bienes o servicios del tipo particular definido por la presente parte.
        :param descripcion: Atributo requerido para precisar la descripción del bien o servicio cubierto por la presente parte.
        :param no_identificacion: Atributo opcional para expresar el número de serie, número de parte del bien o identificador del producto o del servicio amparado por la presente parte. Opcionalmente se puede utilizar claves del estándar GTIN.
        :param unidad: Atributo opcional para precisar la unidad de medida propia de la operación del emisor, aplicable para la cantidad expresada en la parte. La unidad debe corresponder con la descripción de la parte.
        :param valor_unitario: Atributo opcional para precisar el valor o precio unitario del bien o servicio cubierto por la presente parte. No se permiten valores negativos.
        :param importe: Atributo opcional para precisar el importe total de los bienes o servicios de la presente parte. Debe ser equivalente al resultado de multiplicar la cantidad por el valor unitario expresado en la parte. No se permiten valores negativos.
        :param informacion_aduanera: Nodo opcional para introducir la información aduanera aplicable cuando se trate de ventas de primera mano de mercancías importadas o se trate de operaciones de comercio exterior con bienes o servicios.
        """

        super().__init__({
            'ClaveProdServ': clave_prod_serv,
            'Cantidad': cantidad,
            'Descripcion': descripcion,
            'NoIdentificacion': no_identificacion,
            'Unidad': unidad,
            'ValorUnitario': valor_unitario,
            'Importe': importe,
            'InformacionAduanera': informacion_aduanera,
        })


class Concepto(XElement):
    """
    http://www.sat.gob.mx/cfd/3
    Nodo requerido para registrar la información detallada de un bien o servicio amparado en el comprobante.
    """

    def __init__(
            self,
            clave_prod_serv: str,
            cantidad: Decimal | int,
            clave_unidad: str,
            descripcion: str,
            valor_unitario: Decimal | int,
            no_identificacion: str = None,
            unidad: str = None,
            descuento: Decimal | int = None,
            informacion_aduanera: str | Sequence[str] = None,
            cuenta_predial: str = None,
            complemento_concepto: Sequence[CFDI] = None,
            parte: Sequence[Parte | dict] = None,
            traslados: Impuesto | str | Sequence[Impuesto | str | dict] = None,
            retenciones: Impuesto | str | Sequence[Impuesto | str | dict] = None,
            _traslados_incluidos: bool = False
    ):
        """
        Nodo requerido para registrar la información detallada de un bien o servicio amparado en el comprobante.
        
        :param clave_prod_serv: Atributo requerido para expresar la clave del producto o del servicio amparado por el presente concepto. Es requerido y deben utilizar las claves del catálogo de productos y servicios, cuando los conceptos que registren por sus actividades correspondan con dichos conceptos.
        :param cantidad: Atributo requerido para precisar la cantidad de bienes o servicios del tipo particular definido por el presente concepto.
        :param clave_unidad: Atributo requerido para precisar la clave de unidad de medida estandarizada aplicable para la cantidad expresada en el concepto. La unidad debe corresponder con la descripción del concepto.
        :param descripcion: Atributo requerido para precisar la descripción del bien o servicio cubierto por el presente concepto.
        :param valor_unitario: Atributo requerido para precisar el valor o precio unitario del bien o servicio cubierto por el presente concepto.
        :param no_identificacion: Atributo opcional para expresar el número de parte, identificador del producto o del servicio, la clave de producto o servicio, SKU o equivalente, propia de la operación del emisor, amparado por el presente concepto. Opcionalmente se puede utilizar claves del estándar GTIN.
        :param unidad: Atributo opcional para precisar la unidad de medida propia de la operación del emisor, aplicable para la cantidad expresada en el concepto. La unidad debe corresponder con la descripción del concepto.
        :param descuento: Atributo opcional para representar el importe de los descuentos aplicables al concepto. No se permiten valores negativos.
        :param informacion_aduanera: Nodo opcional para introducir la información aduanera aplicable cuando se trate de ventas de primera mano de mercancías importadas o se trate de operaciones de comercio exterior con bienes o servicios.
        :param cuenta_predial: Nodo opcional para asentar el número de cuenta predial con el que fue registrado el inmueble, en el sistema catastral de la entidad federativa de que trate, o bien para incorporar los datos de identificación del certificado de participación inmobiliaria no amortizable.
        :param complemento_concepto: Nodo opcional donde se incluyen los nodos complementarios de extensión al concepto definidos por el SAT, de acuerdo con las disposiciones particulares para un sector o actividad específica.
        :param parte: Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el comprobante fiscal digital por Internet.
        :param traslados: Traslados a aplicar.
        :param retenciones: Retenciones a aplicar.
        :param _traslados_incluidos: si el valor valor_unitario ya incluye traslados.
        """

        super().__init__({
            'ClaveProdServ': clave_prod_serv,
            'Cantidad': cantidad,
            'ClaveUnidad': clave_unidad,
            'Descripcion': descripcion,
            'ValorUnitario': valor_unitario,
            'NoIdentificacion': no_identificacion,
            'Unidad': unidad,
            'Descuento': descuento,
            'InformacionAduanera': informacion_aduanera,
            'CuentaPredial': cuenta_predial,
            'ComplementoConcepto': complemento_concepto,
            'Parte': parte,
            'Impuestos': {
                'Traslados': traslados,
                'Retenciones': retenciones
            },
            '_traslados_incluidos': _traslados_incluidos
        })


class Receptor(XElement):
    """
    http://www.sat.gob.mx/cfd/3
    Nodo requerido para precisar la información del contribuyente receptor del comprobante.
    """

    def __init__(
            self,
            rfc: str,
            uso_cfdi: str,
            nombre: str = None,
            residencia_fiscal: str = None,
            num_reg_id_trib: str = None,
    ):
        """
        Nodo requerido para precisar la información del contribuyente receptor del comprobante.
        
        :param rfc: Atributo requerido para precisar la Clave del Registro Federal de Contribuyentes correspondiente al contribuyente receptor del comprobante.
        :param uso_cfdi: Atributo requerido para expresar la clave del uso que dará a esta factura el receptor del CFDI.
        :param nombre: Atributo opcional para precisar el nombre, denominación o razón social del contribuyente receptor del comprobante.
        :param residencia_fiscal: Atributo condicional para registrar la clave del país de residencia para efectos fiscales del receptor del comprobante, cuando se trate de un extranjero, y que es conforme con la especificación ISO 3166-1 alpha-3. Es requerido cuando se incluya el complemento de comercio exterior o se registre el atributo NumRegIdTrib.
        :param num_reg_id_trib: Atributo condicional para expresar el número de registro de identidad fiscal del receptor cuando sea residente en el extranjero. Es requerido cuando se incluya el complemento de comercio exterior.
        """

        super().__init__({
            'Rfc': rfc,
            'UsoCFDI': uso_cfdi,
            'Nombre': nombre,
            'ResidenciaFiscal': residencia_fiscal,
            'NumRegIdTrib': num_reg_id_trib,
        })


class Comprobante(cfdi40.Comprobante):
    """
    Estándar de Comprobante Fiscal Digital por Internet.
    """
    tag = '{http://www.sat.gob.mx/cfd/3}Comprobante'
    version = '3.3'
    complemento_pago = pago10.Pagos



import base64
import io
from zipfile import ZipFile

from lxml.etree import QName
from packaging import version
import requests

from . import PAC, Environment, Accept, Document, CancelReason, CancelationAcknowledgment
from .. import __version__, ResponseError, CFDI, iterate, Signer, DocumentNotFoundError


def _process_content(response, fmt):
    if content := response.get("content"):
        match fmt:
            case "xml":
                response["xml"] = base64.b64decode(content)
            case "pdf":
                response["pdf"] = base64.b64decode(content)
            case "zip":
                zip_data = base64.b64decode(content)
                with io.BytesIO(zip_data) as b:
                    with ZipFile(b, "r") as zf:
                        response["xml"] = zf.read('invoice.xml')
                        response["pdf"] = zf.read('invoice.pdf')


def _process_format(accept: Accept):
    match accept:
        case Accept.XML_PDF:
            return "zip"
        case Accept.PDF:
            return "pdf"
        case Accept.XML:
            return "xml"
        case _:
            raise NotImplementedError()


class Diverza(PAC):
    """
    Class to call Diverza API's
    see https://diverza.docs.apiary.io/ for more information
    """
    RCF = "SNF171020F3A"

    def __init__(self, rfc: str, id: str, token: str, environment=Environment.PRODUCTION):
        """
        :param rfc: Document emitter's RFC Info
        :param id: ID assigned by Diverza when the client is registered
        :param token: Security token assigned by Diverza. It must match with the Client & Tax ID (RFC)
        """
        super().__init__(environment)
        self.payload_append = {
            "credentials": {
                "id": id,
                "token": token
            },
            "issuer": {
                "rfc": rfc
            }
        }

    def _request(self, path, json, method="post"):
        json = self.payload_append | json
        match self.environment:
            case Environment.PRODUCTION:
                host = "https://servicios.diverza.com"
            case Environment.TEST:
                host = "https://serviciosdemo.diverza.com"
            case _:
                raise NotImplementedError("Environment not Supported")

        r = requests.request(
            method=method,
            url=f"{host}/{path}",
            headers={
                "User-Agent": __version__.__user_agent__
                # Nota: Header de Accept genera errores
            },
            json=json
        )

        if r.ok:
            return r
        else:
            raise ResponseError(r)

    def _issue_stamp(self, cfdi: CFDI, accept: Accept = Accept.XML, ref_id: str = None, operation: str = "stamp") -> Document:
        raw_cfdi = base64.b64encode(cfdi.xml_bytes()).decode()
        fmt = _process_format(accept)

        def _process_vnd():
            if comp := [QName(x.tag).localname for x in iterate(cfdi.get("Complemento"))]:
                if "Nomina" in comp:
                    return f"application/vnd.diverza.cfdi_{cfdi['Version']}_complemento_nomina+xml"
                else:
                    return f"application/vnd.diverza.cfdi_{cfdi['Version']}_complemento+xml"
            else:
                return f"application/vnd.diverza.cfdi_{cfdi['Version']}+xml"

        def _process_path():
            if version.parse(cfdi['Version']) >= version.parse("4.0"):
                return f'api/v2/documents/{operation}'
            return f"api/v1/documents/{operation}"

        json = {
            "issuer": {
                "rfc": cfdi["Emisor"]["Rfc"]
            },
            "receiver": {
                "emails": []
            },
            "document": {
                "certificate-number": cfdi["NoCertificado"],
                "section": "all",
                "format": fmt,
                "type": _process_vnd(),
                "content": raw_cfdi,
                "template": "letter"
            }
        }
        if ref_id:
            json["document"]["ref-id"] = ref_id

        res = self._request(
            path=_process_path(),
            json=json
        ).json()

        _process_content(res, fmt)
        return Document(
            document_id=res['uuid'],
            xml=res.get('xml'),
            pdf=res.get('pdf'),
        )

    def issue(self, cfdi: CFDI, accept: Accept = Accept.XML, ref_id: str = None) -> Document:
        return self._issue_stamp(cfdi=cfdi, accept=accept, ref_id=ref_id, operation="issue")

    def stamp(self, cfdi: CFDI, accept: Accept = Accept.XML, ref_id: str = None) -> Document:
        return self._issue_stamp(cfdi=cfdi, accept=accept, ref_id=ref_id, operation="stamp")

    def cancel(self, cfdi: CFDI, reason: CancelReason, substitution_id: str = None, signer: Signer = None) -> CancelationAcknowledgment:
        def _process_path():
            if version.parse(cfdi['Version']) >= version.parse("4.0"):
                return f"api/v2/documents/{document_id}/cancel"
            return f"api/v1/documents/{document_id}/cancel"

        document_id = cfdi["Complemento"]["TimbreFiscalDigital"]["UUID"]
        res = self._request(
            path=_process_path(),
            json={
                "issuer": {
                    "rfc": cfdi["Emisor"]["Rfc"]
                }
            },
            method="put"
        ).json()

        if res["status"] == "canceled":
            return CancelationAcknowledgment(
                code="202"
            )

        raise Exception("Unknow Response: " + res.text)

    def recover(self, document_id: str, accept: Accept = Accept.XML) -> Document:
        fmt = _process_format(accept)

        def doc_recover(version):
            return self._request(
                path=f"api/{version}/documents/{document_id}/reprint",
                json={
                    "receiver": {
                        "emails": []
                    },
                    "document": {
                        "section": "all",
                        "format": fmt,
                        "template": "letter"
                    }
                }
            ).json()

        try:
            res = doc_recover("v2")
        except ResponseError as ex_v2:
            if ex_v2.response.status_code != 404:
                raise

            try:
                res = doc_recover("v1")
            except ResponseError:
                raise DocumentNotFoundError(ex_v2.response)

        _process_content(res, fmt)
        return Document(
            document_id=res['uuid'],
            xml=res.get('xml'),
            pdf=res.get('pdf'),
        )

    def rfc_valid(self, rfc: str | list[str]) -> bool | list[bool]:
        r = self._request(
            path="api/v1/rfc/status",
            json={
                "identifiers": [
                    {
                        "rfc": r
                    }
                    for r in iterate(rfc)
                ]
            }
        )

        res = r.json()
        if isinstance(rfc, str):
            return res['identifiers'][0]['active'] == "Y"

        return [
            i["active"] == "Y" for i in res['identifiers']
        ]
